import warnings
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import numpy

from . import _compatibility
from .box import Box
from .topology import Angles, Bonds, Dihedrals, Impropers, LabelMap


class Snapshot:
    """Particle configuration.

    A `Snapshot` holds the data for `N` particles, the simulation `Box`, and the
    timestep.

    Parameters
    ----------
    N : int
        Number of particles in configuration.
    box : `Box`
        Simulation box.
    step : int
        Simulation time step. Default of ``None`` means time step is not
        specified.
    num_types : int
        Number of particle types. If ``None``, the number of types is deduced
        from `typeid`.

    Example
    -------
    Here is a 3-particle configuration in a cubic box centered at the
    origin at step 10:

    .. code-block:: python

        snapshot = lammpsio.Snapshot(
            N=3,
            box=lammpsio.Box([-5, -5, -5], [5, 5, 5]),
            step=10
        )

    All values of indexes follow the LAMMPS 1-indexed convention, but the arrays
    themselves are 0-indexed. `Snapshot` will lazily initialize these
    per-particle arrays as they are accessed to save memory. Hence, accessing a
    per-particle property will allocate it to default values. If you want to
    check if an attribute has been set, use the corresponding ``has_`` method
    instead (e.g., `has_position` to check if the `position` data is allocated):

    .. code-block:: python

        snapshot.position = [[0, 0, 0], [1, -1, 1], [1.5, 2.5, -3.5]]
        snapshot.typeid[2] = 2
        if not snapshot.has_mass():
            snapshot.mass = [2., 2., 10.]

    """

    def __init__(
        self,
        N: int,
        box: "Box",
        step: Optional[int] = None,
        num_types: Optional[int] = None,
    ) -> None:
        self._N = N
        self._box = Box.cast(box)
        self.step = step

        self._id = None
        self._position = None
        self._velocity = None
        self._image = None
        self._molecule = None
        self._num_types = None
        self._typeid = None
        self._charge = None
        self._mass = None
        self._type_label = None
        self._bonds = None
        self._angles = None
        self._dihedrals = None
        self._impropers = None

    @classmethod
    def from_hoomd_gsd(
        cls: type["Snapshot"], frame: Any
    ) -> Tuple["Snapshot", Optional[Dict]]:
        """Create from a HOOMD GSD frame.

        Parameters
        ----------
        frame : `gsd.hoomd.Frame`
            HOOMD GSD frame to convert.

        Returns
        -------
        `Snapshot`
            Snapshot created from HOOMD GSD frame.
        dict
            A map from the :attr:`Snapshot.typeid` to the HOOMD type.

            .. deprecated:: 0.7.0
                Use `Snapshot.type_label` instead.

        Example
        -------
        Create snapshot from a GSD file:

        .. skip: start if(gsd == None, reason="gsd not installed")

        .. code-block:: python

            frame = gsd.hoomd.Frame()
            frame.configuration.box = [10, 10, 10, 0, 0, 0]
            frame.particles.N = 3
            snapshot, type_map = lammpsio.Snapshot.from_hoomd_gsd(frame)

        This creates a `Snapshot` from the provided HOOMD GSD frame as well as
        the corresponding dictionary of typeids mapped from the HOOMD GSD frame.

        """
        # ensures frame is well formed and that we have NumPy arrays
        frame.validate()

        # process HOOMD box to LAMMPS box
        box = numpy.array(frame.configuration.box, copy=True)
        box = Box.from_hoomd_convention(box)

        snap = Snapshot(
            N=frame.particles.N,
            box=box,
            step=frame.configuration.step,
        )
        if frame.particles.position is not None:
            snap.position = frame.particles.position

        if frame.particles.velocity is not None:
            snap.velocity = frame.particles.velocity

        if frame.particles.image is not None:
            snap.image = frame.particles.image

        if frame.particles.typeid is not None:
            snap.typeid = frame.particles.typeid + 1

        if frame.particles.charge is not None:
            snap.charge = frame.particles.charge

        if frame.particles.mass is not None:
            snap.mass = frame.particles.mass

        if frame.particles.body is not None:
            snap.molecule = frame.particles.body + 1
            if numpy.any(snap.molecule < 0):
                warnings.warn("Some molecule IDs are negative, remapping needed.")

        # set particle label
        label_map_particle = None
        if frame.particles.types is not None:
            label_map_particle = {
                typeid + 1: i for typeid, i in enumerate(frame.particles.types)
            }
            snap.type_label = LabelMap(map=label_map_particle)

        if (
            frame.bonds.N != 0
            or frame.bonds.group is not None
            or frame.bonds.typeid is not None
            or frame.bonds.types is not None
        ):
            # always create a data container even if there are no bonds
            snap.bonds = Bonds(N=frame.bonds.N)

            if frame.bonds.group is not None:
                snap.bonds.members = frame.bonds.group + 1

            if frame.bonds.typeid is not None:
                snap.bonds.typeid = frame.bonds.typeid + 1

            if frame.bonds.types is not None:
                label_map_bond = {
                    typeid + 1: i for typeid, i in enumerate(frame.bonds.types)
                }
                snap.bonds.type_label = LabelMap(map=label_map_bond)

        if (
            frame.angles.N != 0
            or frame.angles.group is not None
            or frame.angles.typeid is not None
            or frame.angles.types is not None
        ):
            # always create a data container even if there are no angles
            snap.angles = Angles(N=frame.angles.N)

            if frame.angles.group is not None:
                snap.angles.members = frame.angles.group + 1

            if frame.angles.typeid is not None:
                snap.angles.typeid = frame.angles.typeid + 1

            if frame.angles.types is not None:
                label_map_angle = {
                    typeid + 1: i for typeid, i in enumerate(frame.angles.types)
                }
                snap.angles.type_label = LabelMap(map=label_map_angle)

        if (
            frame.dihedrals.N != 0
            or frame.dihedrals.group is not None
            or frame.dihedrals.typeid is not None
            or frame.dihedrals.types is not None
        ):
            # always create a data container even if there are no dihedrals
            snap.dihedrals = Dihedrals(N=frame.dihedrals.N)

            if frame.dihedrals.group is not None:
                snap.dihedrals.members = frame.dihedrals.group + 1

            if frame.dihedrals.typeid is not None:
                snap.dihedrals.typeid = frame.dihedrals.typeid + 1

            if frame.dihedrals.types is not None:
                label_map_dihedral = {
                    typeid + 1: i for typeid, i in enumerate(frame.dihedrals.types)
                }
                snap.dihedrals.type_label = LabelMap(map=label_map_dihedral)

        if (
            frame.impropers.N != 0
            or frame.impropers.group is not None
            or frame.impropers.typeid is not None
            or frame.impropers.types is not None
        ):
            # always create a data container even if there are no impropers
            snap.impropers = Impropers(N=frame.impropers.N)

            if frame.impropers.group is not None:
                snap.impropers.members = frame.impropers.group + 1

            if frame.impropers.typeid is not None:
                snap.impropers.typeid = frame.impropers.typeid + 1

            if frame.impropers.types is not None:
                label_map_improper = {
                    typeid + 1: i for typeid, i in enumerate(frame.impropers.types)
                }
                snap.impropers.type_label = LabelMap(map=label_map_improper)

        return snap, label_map_particle

    def to_hoomd_gsd(self, type_map: Optional[Dict[int, Any]] = None) -> Any:
        """Create a HOOMD GSD frame.

        Parameters
        ----------
        type_map : dict
            A map from the :attr:`Snapshot.typeid` to a HOOMD type.
            If not specified, the typeids are used as the types.

            .. deprecated:: 0.7.0
                Use `Snapshot.type_label` instead.

        Returns
        -------
        `gsd.hoomd.Frame`
            Converted HOOMD GSD frame.

        Example
        -------

        Convert snapshot to a GSD file:

        .. code-block:: python

            frame = snapshot.to_hoomd_gsd()

        This creates a GSD frame for the HOOMD-blue simulation package
        from the `Snapshot` object.

        .. skip: end

        """
        if _compatibility.gsd_version is None:
            raise ImportError("GSD package not found")

        frame = _compatibility.gsd_frame_class()

        if self.step is not None:
            frame.configuration.step = int(self.step)

        frame.configuration.box = self.box.to_hoomd_convention()

        # sort tags if specified and not increasing because GSD guarantees an order
        reverse_order = None
        if self.has_id() and self.N > 1 and not numpy.all(self.id[1:] > self.id[:-1]):
            order = numpy.argsort(self.id)
            self.reorder(order, check_order=False)
            # build the reverse map to undo the sort later
            reverse_order = numpy.zeros(self.N, dtype=int)
            for i, v in enumerate(order):
                reverse_order[v] = i

        frame.particles.N = self.N
        if self.has_position():
            # Center the positions using HOOMD tilt factors (computed above)
            low, matrix = self.box.to_matrix()
            center = low + 0.5 * numpy.sum(matrix, axis=1)
            frame.particles.position = self.position - center
        if self.has_velocity():
            frame.particles.velocity = self.velocity.copy()
        if self.has_image():
            frame.particles.image = self.image.copy()
        if type_map is not None:
            warnings.warn(
                "type_map is deprecated, use Snapshot.type_label instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            type_label_map = LabelMap(map=type_map)
        else:
            type_label_map = self.type_label
        if self.has_typeid():
            frame.particles.typeid = numpy.zeros(self.N, dtype=int)
            type_label_map = _set_type_id(
                self.typeid, frame.particles.typeid, type_label_map
            )
            frame.particles.types = type_label_map.types
        if self.has_charge():
            frame.particles.charge = self.charge.copy()
        if self.has_mass():
            frame.particles.mass = self.mass.copy()
        if self.has_molecule():
            frame.particles.body = self.molecule - 1

        if self.bonds is not None:
            frame.bonds.N = self.bonds.N
            if self.bonds.has_members():
                frame.bonds.group = self.bonds.members - 1
            bond_label_map = self.bonds.type_label
            if self.bonds.has_typeid():
                frame.bonds.typeid = numpy.zeros(self.bonds.N, dtype=int)
                bond_label_map = _set_type_id(
                    self.bonds.typeid,
                    frame.bonds.typeid,
                    bond_label_map,
                )
            frame.bonds.types = bond_label_map.types

        if self.angles is not None:
            frame.angles.N = self.angles.N
            if self.angles.has_members():
                frame.angles.group = self.angles.members - 1
            angle_label_map = self.angles.type_label
            if self.angles.has_typeid():
                frame.angles.typeid = numpy.zeros(self.angles.N, dtype=int)
                angle_label_map = _set_type_id(
                    self.angles.typeid,
                    frame.angles.typeid,
                    angle_label_map,
                )
            frame.angles.types = angle_label_map.types

        if self.dihedrals is not None:
            frame.dihedrals.N = self.dihedrals.N
            if self.dihedrals.has_members():
                frame.dihedrals.group = self.dihedrals.members - 1
            dihedral_label_map = self.dihedrals.type_label
            if self.dihedrals.has_typeid():
                frame.dihedrals.typeid = numpy.zeros(self.dihedrals.N, dtype=int)
                dihedral_label_map = _set_type_id(
                    self.dihedrals.typeid,
                    frame.dihedrals.typeid,
                    dihedral_label_map,
                )
            frame.dihedrals.types = dihedral_label_map.types

        if self.impropers is not None:
            frame.impropers.N = self.impropers.N
            if self.impropers.has_members():
                frame.impropers.group = self.impropers.members - 1
            improper_label_map = self.impropers.type_label
            if self.impropers.has_typeid():
                frame.impropers.typeid = numpy.zeros(self.impropers.N, dtype=int)
                improper_label_map = _set_type_id(
                    self.impropers.typeid,
                    frame.impropers.typeid,
                    improper_label_map,
                )
            frame.impropers.types = improper_label_map.types

        # undo the sort so object goes back the way it was
        if reverse_order is not None:
            self.reorder(reverse_order, check_order=False)

        return frame

    @property
    def N(self) -> int:
        """int: Number of particles."""
        return self._N

    @property
    def box(self) -> "Box":
        """`Box`: Simulation box."""
        return self._box

    @property
    def step(self) -> Optional[int]:
        """int: Simulation time step."""
        return self._step

    @step.setter
    def step(self, value):
        if value is not None:
            self._step = int(value)
        else:
            self._step = None

    @property
    def id(self) -> numpy.ndarray:
        """(*N*,) `numpy.ndarray` of `int`: Particle IDs.

        The default value on initialization runs from 1 to *N*.

        """
        if not self.has_id():
            self._id = numpy.arange(1, self.N + 1)
        return self._id

    @id.setter
    def id(self, value):
        if value is not None:
            v = numpy.array(
                value, ndmin=1, copy=_compatibility.numpy_copy_if_needed, dtype=int
            )
            if v.shape != (self.N,):
                raise TypeError("Ids must be a size N array")
            if not self.has_id():
                self._id = numpy.arange(1, self.N + 1)
            numpy.copyto(self._id, v)
        else:
            self._id = None

    def has_id(self):
        """Check if configuration has particle IDs.

        Returns
        -------
        bool
            ``True`` if particle IDs have been initialized.
        """
        return self._id is not None

    @property
    def position(self) -> numpy.ndarray:
        """(*N*, 3) `numpy.ndarray` of `float`: Positions.

        The default value on initialization is 0 for all entries.

        """
        if not self.has_position():
            self._position = numpy.zeros((self.N, 3), dtype=float)
        return self._position

    @position.setter
    def position(self, value):
        if value is not None:
            v = numpy.array(
                value, ndmin=2, copy=_compatibility.numpy_copy_if_needed, dtype=float
            )
            if v.shape != (self.N, 3):
                raise TypeError("Positions must be an Nx3 array")
            if not self.has_position():
                self._position = numpy.zeros((self.N, 3), dtype=float)
            numpy.copyto(self._position, v)
        else:
            self._position = None

    def has_position(self):
        """Check if configuration has positions.

        Returns
        -------
        bool
            ``True`` if positions have been initialized.

        """
        return self._position is not None

    @property
    def image(self) -> numpy.ndarray:
        """(*N*, 3) `numpy.ndarray` of `int`: Images.

        The default value on initialization is 0 for all entries.

        """
        if not self.has_image():
            self._image = numpy.zeros((self.N, 3), dtype=int)
        return self._image

    @image.setter
    def image(self, value):
        if value is not None:
            v = numpy.array(
                value, ndmin=2, copy=_compatibility.numpy_copy_if_needed, dtype=int
            )
            if v.shape != (self.N, 3):
                raise TypeError("Images must be an Nx3 array")
            if not self.has_image():
                self._image = numpy.zeros((self.N, 3), dtype=int)
            numpy.copyto(self._image, v)
        else:
            self._image = None

    def has_image(self):
        """Check if configuration has images.

        Returns
        -------
        bool
            ``True`` if images have been initialized.
        """
        return self._image is not None

    @property
    def velocity(self) -> numpy.ndarray:
        """(*N*, 3) `numpy.ndarray` of `float`: Velocities.

        The default value on initialization is 0 for all entries.

        """
        if not self.has_velocity():
            self._velocity = numpy.zeros((self.N, 3), dtype=float)
        return self._velocity

    @velocity.setter
    def velocity(self, value):
        if value is not None:
            v = numpy.array(
                value, ndmin=2, copy=_compatibility.numpy_copy_if_needed, dtype=float
            )
            if v.shape != (self.N, 3):
                raise TypeError("Velocities must be an Nx3 array")
            if not self.has_velocity():
                self._velocity = numpy.zeros((self.N, 3), dtype=float)
            numpy.copyto(self._velocity, v)
        else:
            self._velocity = None

    def has_velocity(self):
        """Check if configuration has velocities.

        Returns
        -------
        bool
            ``True`` if velocities have been initialized.
        """
        return self._velocity is not None

    @property
    def molecule(self) -> numpy.ndarray:
        """(*N*,) `numpy.ndarray` of `int`: Molecule tags.

        The default value on initialization is 0 for all entries.

        """
        if not self.has_molecule():
            self._molecule = numpy.zeros(self.N, dtype=int)
        return self._molecule

    @molecule.setter
    def molecule(self, value):
        if value is not None:
            v = numpy.array(
                value, ndmin=1, copy=_compatibility.numpy_copy_if_needed, dtype=int
            )
            if v.shape != (self.N,):
                raise TypeError("Molecules must be a size N array")
            if not self.has_molecule():
                self._molecule = numpy.zeros(self.N, dtype=int)
            numpy.copyto(self._molecule, v)
        else:
            self._molecule = None

    def has_molecule(self):
        """Check if configuration has molecule tags.

        Returns
        -------
        bool
            ``True`` if molecule tags have been initialized.

        """
        return self._molecule is not None

    @property
    def num_types(self) -> int:
        """int: Number of atom types."""
        if self._num_types is not None:
            return self._num_types
        else:
            if self.has_typeid():
                return numpy.amax(self.typeid)
            else:
                return 1

    @num_types.setter
    def num_types(self, value):
        if value is not None:
            self._num_types = int(value)
        else:
            self._num_types = None

    @property
    def typeid(self) -> numpy.ndarray:
        """(*N*,) `numpy.ndarray` of `int`: Types.

        The default value on initialization is 1 for all entries.

        """
        if not self.has_typeid():
            self._typeid = numpy.ones(self.N, dtype=int)
        return self._typeid

    @typeid.setter
    def typeid(self, value):
        if value is not None:
            v = numpy.array(
                value, ndmin=1, copy=_compatibility.numpy_copy_if_needed, dtype=int
            )
            if v.shape != (self.N,):
                raise TypeError("Type must be a size N array")
            if not self.has_typeid():
                self._typeid = numpy.ones(self.N, dtype=int)
            numpy.copyto(self._typeid, v)
        else:
            self._typeid = None

    def has_typeid(self):
        """Check if configuration has types.

        Returns
        -------
        bool
            ``True`` if types have been initialized.
        """
        return self._typeid is not None

    @property
    def charge(self) -> numpy.ndarray:
        """(*N*,) `numpy.ndarray` of `float`: Charges.

        The default value on initialization is 0 for all entries.

        """
        if not self.has_charge():
            self._charge = numpy.zeros(self.N, dtype=float)
        return self._charge

    @charge.setter
    def charge(self, value):
        if value is not None:
            v = numpy.array(
                value, ndmin=1, copy=_compatibility.numpy_copy_if_needed, dtype=float
            )
            if v.shape != (self.N,):
                raise TypeError("Charge must be a size N array")
            if not self.has_charge():
                self._charge = numpy.zeros(self.N, dtype=float)
            numpy.copyto(self._charge, v)
        else:
            self._charge = None

    def has_charge(self):
        """Check if configuration has charges.

        Returns
        -------
        bool
            ``True`` if charges have been initialized.
        """
        return self._charge is not None

    @property
    def mass(self) -> numpy.ndarray:
        """(*N*,) `numpy.ndarray` of `float`: Masses.

        The default value on initialization is 1 for all entries.

        """
        if not self.has_mass():
            self._mass = numpy.ones(self.N, dtype=float)
        return self._mass

    @mass.setter
    def mass(self, value):
        if value is not None:
            v = numpy.array(
                value, ndmin=1, copy=_compatibility.numpy_copy_if_needed, dtype=float
            )
            if v.shape != (self.N,):
                raise TypeError("Mass must be a size N array")
            if not self.has_mass():
                self._mass = numpy.ones(self.N, dtype=float)
            numpy.copyto(self._mass, v)
        else:
            self._mass = None

    def has_mass(self):
        """Check if configuration has masses.

        Returns
        -------
        bool
            ``True`` if masses have been initialized.
        """
        return self._mass is not None

    @property
    def type_label(self) -> Optional["LabelMap"]:
        """`LabelMap`: Labels for `typeid`."""
        return self._type_label

    @type_label.setter
    def type_label(self, value):
        if value is not None:
            if not isinstance(value, LabelMap):
                raise TypeError("type_label must be a LabelMap")
            self._type_label = value
        else:
            self._type_label = None

    @property
    def bonds(self) -> Optional["Bonds"]:
        """`Bonds`: Bond data."""
        return self._bonds

    @bonds.setter
    def bonds(self, value):
        if value is not None:
            if not isinstance(value, Bonds):
                raise TypeError("bonds must be Bonds")
            self._bonds = value
        else:
            self._bonds = None

    def has_bonds(self):
        """Check if configuration has bonds.

        Returns
        -------
        bool
            ``True`` if bonds is initialized and there is at least one bond.
        """
        return self._bonds is not None and self._bonds.N > 0

    @property
    def angles(self) -> Optional["Angles"]:
        """`Angles`: Angle data."""
        return self._angles

    @angles.setter
    def angles(self, value):
        if value is not None:
            if not isinstance(value, Angles):
                raise TypeError("angles must be Angles")
            self._angles = value
        else:
            self._angles = None

    def has_angles(self):
        """Check if configuration has angles.

        Returns
        -------
        bool
            ``True`` if angles is initialized and there is at least one angle.
        """
        return self._angles is not None and self._angles.N > 0

    @property
    def dihedrals(self) -> Optional["Dihedrals"]:
        """`Dihedrals`: Dihedral data."""
        return self._dihedrals

    @dihedrals.setter
    def dihedrals(self, value):
        if value is not None:
            if not isinstance(value, Dihedrals):
                raise TypeError("dihedrals must be Dihedrals")
            self._dihedrals = value
        else:
            self._dihedrals = None

    def has_dihedrals(self):
        """Check if configuration has dihedrals.

        Returns
        -------
        bool
            ``True`` if dihedrals is initialized and there is at least one dihedral.
        """
        return self._dihedrals is not None and self._dihedrals.N > 0

    @property
    def impropers(self) -> Optional["Impropers"]:
        """`Impropers`: Improper data."""
        return self._impropers

    @impropers.setter
    def impropers(self, value):
        if value is not None:
            if not isinstance(value, Impropers):
                raise TypeError("impropers must be Impropers")
            self._impropers = value
        else:
            self._impropers = None

    def has_impropers(self):
        """Check if configuration has impropers.

        Returns
        -------
        bool
            ``True`` if impropers is initialized and there is at least one improper.
        """
        return self._impropers is not None and self._impropers.N > 0

    def reorder(self, order: Any, check_order: bool = True) -> None:
        """Reorder the particles in place.

        Parameters
        ----------
        order : list
            New order of indexes.
        check_order : bool
            If ``True``, validate the new ``order`` before applying it.

        Example
        -------

        Reorder the particles in a snapshot

        .. code-block:: python

            snapshot.bonds = lammpsio.Bonds(N=3, num_types=1)
            bond_id = [1, 3, 2]
            members = [[1, 2],
                       [2, 3],
                       [3, 1]]
            typeid = [1, 1, 1]

            snapshot.bonds.id = bond_id
            snapshot.bonds.typeid = typeid
            snapshot.bonds.members = members

            snapshot.bonds.reorder(numpy.sort(numpy.array(bond_id)) - 1,
                                   check_order=True)

        This reorders the particle data with the same ordering as the bonds.
        To ensure only the right unique indexes are used, the ``check_order``
        parameter is set to True. In LAMMPS, all the IDs are 1-indexed, while,
        Python is 0-indexed. Thus the ``bond_id`` is also decreased by 1
        to match the Python convention.

        """
        # sanity check the sorting order before applying it
        if check_order and self.N > 1:
            sorted_order = numpy.sort(order)
            if (
                sorted_order[0] != 0
                or sorted_order[-1] != self.N - 1
                or not numpy.all(sorted_order[1:] == sorted_order[:-1] + 1)
            ):
                raise ValueError("New order must be an array from 0 to N-1")

        if self.has_id():
            self._id = self._id[order]
        if self.has_position():
            self._position = self._position[order]
        if self.has_velocity():
            self._velocity = self._velocity[order]
        if self.has_image():
            self._image = self._image[order]
        if self.has_molecule():
            self._molecule = self._molecule[order]
        if self.has_typeid():
            self._typeid = self._typeid[order]
        if self.has_charge():
            self._charge = self._charge[order]
        if self.has_mass():
            self._mass = self._mass[order]


def _set_type_id(
    lammps_typeid: Union[Sequence[int], numpy.ndarray],
    gsd_typeid: Union[Sequence[int], numpy.ndarray],
    label_map: Optional["LabelMap"],
) -> "LabelMap":
    """Maps LAMMPS typeids to HOOMD GSD typeids using a given label map.

    Parameters
    ----------
    lammps_typeid : list
        List of LAMMPS typeids to be mapped (one-indexed).
    gsd_typeid : list
        List of HOOMD GSD typeids to be updated (zero-indexed).
    label_map : `LabelMap`
        LabelMap for connection type mapping LAMMPS typeids to HOOMD GSD types.

    Returns
    -------
    `LabelMap`
        LabelMap mapping LAMMPS typeids to HOOMD GSD types.
    """
    if label_map is None:
        sorted_typeids = numpy.sort(numpy.unique(lammps_typeid))
        label_map = {typeid: str(typeid) for typeid in sorted_typeids}
        label_map = LabelMap(map=label_map)

    hoomd_type_map = {
        typeid: typeidx for typeidx, typeid in enumerate(label_map.keys())
    }
    for i, typeid in enumerate(lammps_typeid):
        gsd_typeid[i] = hoomd_type_map[typeid]

    return label_map



from typing import Optional, Sequence, Tuple, Type, Union

import numpy

from . import _compatibility


class Box:
    r"""Simulation box.

    In LAMMPS, the simulation box is specified by three parameters: `low`,
    `high`, and `tilt`. `low` defines the origin (lower corner) of the box,
    while `high` specifies how far the box extends along each axis. The
    difference between `high` and `low` gives three lengths $L_x$, $L_y$, and
    $L_z$. `tilt` has three factors ($L_{xy}$, $L_{xz}$, $L_{yz}$) that skew the
    edges to create non-orthorhombic simulation boxes. These parameters define
    a box matrix consisting of three lattice vectors **a**, **b**, and **c**:

    .. math::

        [\mathbf{a} \quad \mathbf{b} \quad \mathbf{c} ] =
        \begin{bmatrix}
            L_x & L_{xy} & L_{xz} \\
            0 & L_y & L_{yz} \\
            0 & 0 & L_z
        \end{bmatrix}

    For more details on how to convert between the LAMMPS parameters and box
    matrix see the `LAMMPS documentation
    <https://docs.lammps.org/Howto_triclinic.html#transformation-from-general-to-restricted-triclinic-boxes>`_.

    .. warning::

        `high` is the upper bound of the simulation box **only** when it is
        orthorhombic.

    Parameters
    ----------
    low : list
        Origin of the box
    high : list
        High parameter used to compute edge lengths.
    tilt : list
        Tilt factors ``xy``, ``xz``, and ``yz`` for a triclinic box. Default of
        ``None`` is a strictly orthorhombic box, implying all are zero.

    Examples
    --------
    Construct a orthorhombic simulation box with edge lengths (10, 10, 10)
    and origin (-5, -5, 5):

    .. code-block:: python

        box = lammpsio.Box(low=[-5, -5, -5], high=[5, 5, 5])

    The tilt defaults to ``None``, meaning the box is orthorhombic.

    Construct a triclinic simulation box:

    .. code-block:: python

        box = lammpsio.Box(
            low=[-5.0, -5.0, -5.0],
            high=[5.0, 5.0, 5.0],
            tilt=[1.0, -2.0, 0.5]
        )

    This box has the same nominal $L_x$, $L_y$, and $L_z$ as the
    orthorhombic box, but now it also has tilt factors (1, -2, 0.5).

    """

    def __init__(
        self,
        low: Union[Sequence[float], numpy.ndarray],
        high: Union[Sequence[float], numpy.ndarray],
        tilt: Optional[Union[Sequence[float], numpy.ndarray]] = None,
    ) -> None:
        self.low = low
        self.high = high
        self.tilt = tilt

    @classmethod
    def cast(cls: Type["Box"], value: Union[Sequence[float], numpy.ndarray]) -> "Box":
        """Cast from an array.

        If ``value`` has 6 elements, it is unpacked as an orthorhombic box::

            x_lo, y_lo, z_lo, x_hi, y_hi, z_hi = value

        If ``value`` has 9 elements, it is unpacked as a triclinic box::

            x_lo, y_lo, z_lo, x_hi, y_hi, z_hi, xy, xz, yz = value

        Parameters
        ----------
        value : list
            6-element or 9-element array representing the box.

        Returns
        -------
        `Box`
            A simulation box.

        Examples
        --------
        Construct an orthorhombic simulation box by casting an array:

        .. code-block:: python

            box = lammpsio.Box.cast([-5.0, -5.0, -5.0, 5.0, 5.0, 5.0])

        The array defines an orthorhombic box, with (-5.0, -5.0, -5.0) being cast
        to the `low` and (5.0, 5.0, 5.0) being cast to the `high` lists in the
        `Box` format.

        """
        if isinstance(value, Box):
            return value
        v = numpy.array(
            value, ndmin=1, copy=_compatibility.numpy_copy_if_needed, dtype=float
        )
        if v.shape == (9,):
            return Box(v[:3], v[3:6], v[6:])
        elif v.shape == (6,):
            return Box(v[:3], v[3:])
        else:
            raise TypeError(f"Unable to cast boxlike object with shape {v.shape}")

    def to_matrix(self) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """Convert to an origin and matrix.

        Parameters
        ----------
        box : `Box`
            The box to convert.

        Returns
        -------
        list
            Origin of the box.
        `numpy.ndarray`
            Upper triangular matrix in LAMMPS style::

                [[lx, xy, xz],
                 [0, ly, yz],
                 [0, 0, lz]]

        Examples
        --------
        Convert a box to an origin and matrix:

        .. code-block:: python

            low, matrix = box.to_matrix()

        `low` is the origin of the box, and `matrix` is the upper triangular
        matrix that defines the box dimensions and tilt factors.

        """

        low = self.low
        high = self.high
        tilt = self.tilt if self.tilt is not None else [0, 0, 0]

        return (
            low,
            numpy.array(
                [
                    [high[0] - low[0], tilt[0], tilt[1]],
                    [0, high[1] - low[1], tilt[2]],
                    [0, 0, high[2] - low[2]],
                ]
            ),
        )

    @classmethod
    def from_matrix(
        cls: Type["Box"],
        low: Union[Sequence[float], numpy.ndarray],
        matrix: Union[numpy.ndarray, Sequence[Sequence[float]]],
        force_triclinic: bool = False,
    ) -> "Box":
        """Cast from an origin and matrix.

        Parameters
        ----------
        low : list
            Origin of the box.
        matrix : `numpy.ndarray`
            Box matrix.
        force_triclinic : bool
            If ``True``, forces the box to be triclinic even if the tilt
            factors are zero. Default is ``False``.

        Returns
        -------
        `Box`
            A simulation box.

        Examples
        --------

        Construct an orthorhombic simulation box from a low and matrix:

        .. code-block:: python

            box = lammpsio.Box.from_matrix(
                low=[0, 0, 0],
                matrix=[
                    [1, 0, 0],
                    [0, 1, 0],
                    [0, 0, 1]
                ],
                force_triclinic=True)

        If the tilt factors in the matrix are set to zero,
        the method sets the tilt factors to `None` by default.
        By setting the ``force_triclinic`` to `True`,
        the tilt factors are set to (0, 0, 0).

        Construct a triclinic simulation box from a low and matrix:

        .. code-block:: python

            box = lammpsio.Box.from_matrix(
                low=[0, 0, 0],
                matrix=[
                    [1, 1.0, -2.0],
                    [0, 1, 0.5],
                    [0, 0, 1]
                ])

        This creates a triclinic box of unit length in each direction with origin
        at (0, 0, 0) and the tilt factors (1.0, -2.0, 0.5).

        """
        low = numpy.array(low, dtype=float)
        arr = numpy.array(matrix, dtype=float)

        if low.shape != (3,):
            raise TypeError("Low must be a 3-tuple")
        if arr.shape != (3, 3):
            raise TypeError("Box matrix must be a 3x3 array")
        if arr[1, 0] != 0 or arr[2, 0] != 0 or arr[2, 1] != 0:
            raise ValueError("Box matrix must be upper triangular")

        # Calculate high from the matrix
        high = low + numpy.diag(arr)

        # Extract tilt factors
        tilt = [arr[0, 1], arr[0, 2], arr[1, 2]]
        if not force_triclinic and not numpy.any(tilt):
            tilt = None

        return cls(low, high, tilt)

    def to_hoomd_convention(self) -> numpy.ndarray:
        """Convert to HOOMD-blue convention.

        This convention for defining the box is based on
        `HOOMD-blue <https://hoomd-blue.readthedocs.io/en/v5.0.1/hoomd/box.html>`_.

        Parameters
        ----------
        box : `Box`
            The box to convert.

        Returns
        -------
        :class:`numpy.ndarray`
            A matrix of box dimensions in HOOMD-blue convention.

        Examples
        --------
        Convert a LAMMPS simulation box to HOOMD-blue convention:

        .. code-block:: python

            box = lammpsio.Box([-5, -5, -5], [5, 5, 5]).to_hoomd_convention()

        This creates an orthorhombic box of dimensions (10, 10, 10) with `low`
        at (-5, -5, -5) and center of the box at (0, 0, 0).
        In accordance with the HOOMD-blue convention, the tilt factors are
        either normalized by $L_{y}$ for $L_{xy}$ tilt factor and $L_{z}$ for
        $L_{yz}$ and $L_{xz}$ tilt factors. If tilt factors are `None`,
        they are set to zero.

        """

        L = self.high - self.low
        if self.tilt is not None:
            tilt = self.tilt.copy()
            tilt[0] /= L[1]
            tilt[1:] /= L[2]
        else:
            tilt = [0, 0, 0]

        return numpy.concatenate((L, tilt))

    @classmethod
    def from_hoomd_convention(
        cls: Type["Box"],
        box_data: numpy.ndarray,
        low: Optional[Union[Sequence[float], numpy.ndarray]] = None,
        force_triclinic: bool = False,
        dimensions: Optional[int] = None,
    ) -> "Box":
        """Cast from HOOMD-blue convention.

        Parameters
        ----------
        box_data : list
            An array of box dimensions in HOOMD-blue convention.
        low : list
            Origin of the box. If ``None``, the box is centered at the origin.
            Default is ``None``.
        force_triclinic : bool
            If ``True``, forces the box to be triclinic even if the tilt
            factors are zero. Default is ``False``.
        dimensions : int
            The number of dimensions of the box. If ``None``, it is inferred
            from the box data. Default is ``None``.

        Returns
        -------
        box : `Box`
            A simulation box in LAMMPS convention.

        Examples
        --------
        Convert a HOOMD-blue simulation box to the LAMMPS convention:

        .. code-block:: python

            hoomd_box = numpy.array([10, 10, 10, 0, 0, 0])
            lammps_box = lammpsio.Box.from_hoomd_convention(
                box_data=hoomd_box,
                low=[0, 0, 0]
            )

        This creates a orthorhombic box of dimensions  $L_x = 10, L_y = 10$
        and $L_z = 10$ with tilt factors set to (0, 0, 0).
        If `low` is `None`, the box is centered at (0, 0, 0). However, since `low` is
        set to (0, 0, 0), the box is centered at (5, 5, 5).
        The tilt factors are multiplied by $L_y$ for $L_{xy}$ and $L_z$
        for $L_{xz}$ and $L_{yz}$ to convert them to the LAMMPS convention.

        .. note::
            $L_z$ is changed to one if the input $L_z$ is zero,
            as LAMMPS does not allow zero height box.

        """

        if box_data.shape != (6,):
            raise TypeError("Box data must be a 6-tuple")
        if dimensions is None:
            dimensions = 3 if box_data[2] != 0 else 2
        if dimensions not in (2, 3):
            raise ValueError("Dimensions must be 2 or 3")

        L = box_data[:3]
        tilt = box_data[3:]
        if dimensions == 3:
            tilt[0] *= L[1]
            tilt[1:] *= L[2]
        elif dimensions == 2:
            tilt[0] *= L[1]
            tilt[1] = 0
            tilt[2] = 0
            # HOOMD boxes can have Lz = 0, but LAMMPS does not allow this.
            if L[2] == 0:
                L[2] = 1.0

        matrix = numpy.array(
            [[L[0], tilt[0], tilt[1]], [0, L[1], tilt[2]], [0, 0, L[2]]]
        )

        # center the box if low is not provided
        if low is None:
            low = -0.5 * numpy.sum(matrix, axis=1)

        return cls.from_matrix(low, matrix, force_triclinic)

    @property
    def low(self) -> numpy.ndarray:
        """(3,) `numpy.ndarray` of `float`: Low parameter.

        The low of the box is the origin.
        """
        return self._low

    @low.setter
    def low(self, value: Union[Sequence[float], numpy.ndarray]) -> None:
        v = numpy.array(value, ndmin=1, copy=True, dtype=float)
        if v.shape != (3,):
            raise TypeError("Low must be a 3-tuple")
        self._low = v

    @property
    def high(self) -> numpy.ndarray:
        """(3,) `numpy.ndarray` of `float`: High parameter.

        The high of the box is used to compute the lengths $L_x$, $L_y$, and
        $L_z$.
        """
        return self._high

    @high.setter
    def high(self, value: Union[Sequence[float], numpy.ndarray]) -> None:
        v = numpy.array(value, ndmin=1, copy=True, dtype=float)
        if v.shape != (3,):
            raise TypeError("High must be a 3-tuple")
        self._high = v

    @property
    def tilt(self) -> Optional[numpy.ndarray]:
        """(3,) `numpy.ndarray` of `float`: Tilt parameters.

        The 3 tilt factors, $L_{xy}$, $L_{xz}$, and $L_{yz}$ define the
        shape of the box. The default of ``None`` is strictly orthorhombic,
        meaning all are zero.
        """
        return self._tilt

    @tilt.setter
    def tilt(self, value: Optional[Union[Sequence[float], numpy.ndarray]]) -> None:
        v = value
        if v is not None:
            v = numpy.array(v, ndmin=1, copy=True, dtype=float)
            if v.shape != (3,):
                raise TypeError("Tilt must be a 3-tuple")
        self._tilt = v



import numpy
import pytest
from pytest_lazy_fixtures import lf

import lammpsio


def test_orthorhombic(orthorhombic):
    box = orthorhombic
    assert numpy.allclose(box.low, [-5, -10, -1])
    assert numpy.allclose(box.high, [2, 10, 8])
    assert box.tilt is None

    box.low = [0, 0, 0]
    assert numpy.allclose(box.low, [0, 0, 0])
    with pytest.raises(TypeError):
        box.low = [0, 0]

    box.high = [6, 20, 8]
    assert numpy.allclose(box.high, [6, 20, 8])
    with pytest.raises(TypeError):
        box.high = [0, 0]


def test_triclinic(triclinic):
    box = triclinic
    assert numpy.allclose(box.low, [-5, -10, -1])
    assert numpy.allclose(box.high, [2, 10, 8])
    assert numpy.allclose(box.tilt, [2.0, -2.0, 0.5])

    box.low = [0, 0, 0]
    assert numpy.allclose(box.low, [0, 0, 0])
    with pytest.raises(TypeError):
        box.low = [0, 0]

    box.high = [6, 20, 8]
    assert numpy.allclose(box.high, [6, 20, 8])
    with pytest.raises(TypeError):
        box.high = [0, 0]

    box.tilt = [0, 0, 0]
    assert numpy.allclose(box.tilt, [0, 0, 0])
    with pytest.raises(TypeError):
        box.tilt = [0, 0]


@pytest.mark.parametrize("box", [lf("orthorhombic"), lf("triclinic")])
@pytest.mark.parametrize("force_triclinic", [True, False])
def test_from_matrix(box, force_triclinic):
    lx, ly, lz = box.high - box.low
    xy, xz, yz = box.tilt if box.tilt is not None else (0, 0, 0)
    matrix = numpy.array([[lx, xy, xz], [0, ly, yz], [0, 0, lz]])
    new_box = lammpsio.Box.from_matrix(box.low, matrix, force_triclinic=force_triclinic)

    assert numpy.allclose(new_box.low, box.low)
    assert numpy.allclose(new_box.high, box.high)
    if force_triclinic:
        assert new_box.tilt is not None
        if box.tilt is not None:
            assert numpy.allclose(new_box.tilt, box.tilt)
        else:
            assert numpy.allclose(new_box.tilt, [0, 0, 0])
    else:
        if box.tilt is not None:
            assert numpy.allclose(new_box.tilt, box.tilt)
        else:
            assert new_box.tilt is None

    # test with invalid low
    with pytest.raises(TypeError):
        lammpsio.Box.from_matrix([0, 0], matrix)

    # test with invalid matrix shape
    invalid_matrix_shape = numpy.array([[lx, xy], [ly, yz]])
    with pytest.raises(TypeError):
        lammpsio.Box.from_matrix(box.low, invalid_matrix_shape)

    # test with invalid matrix values
    invalid_matrix = numpy.array([[lx, xy, xz], [ly, 0, yz], [lz, 0, 0]])
    with pytest.raises(ValueError):
        lammpsio.Box.from_matrix(box.low, invalid_matrix)


@pytest.mark.parametrize("box", [lf("orthorhombic"), lf("triclinic")])
def test_to_matrix(box):
    low, matrix = box.to_matrix()

    lx, ly, lz = box.high - box.low
    xy, xz, yz = box.tilt if box.tilt is not None else (0, 0, 0)
    expected_matrix = numpy.array([[lx, xy, xz], [0, ly, yz], [0, 0, lz]])

    assert numpy.allclose(matrix, expected_matrix)
    assert numpy.allclose(low, box.low)


@pytest.mark.parametrize("box", [lf("orthorhombic"), lf("triclinic")])
def test_convert_convention(box):
    # Convert to HOOMD-blue convention
    hoomd_box = box.to_hoomd_convention()
    assert isinstance(hoomd_box, numpy.ndarray)
    assert hoomd_box.shape == (6,)
    assert numpy.allclose(hoomd_box[:3], box.high - box.low)
    if box.tilt is not None:
        assert numpy.allclose(
            hoomd_box[3:],
            [
                box.tilt[0] / (box.high[1] - box.low[1]),
                box.tilt[1] / (box.high[2] - box.low[2]),
                box.tilt[2] / (box.high[2] - box.low[2]),
            ],
        )
    else:
        assert numpy.allclose(hoomd_box[3:], [0, 0, 0])

    # Convert back from HOOMD-blue convention
    new_box = lammpsio.Box.from_hoomd_convention(hoomd_box, low=box.low)
    assert numpy.allclose(new_box.low, box.low)
    assert numpy.allclose(new_box.high, box.high)
    if box.tilt is not None:
        assert numpy.allclose(new_box.tilt, box.tilt)
    else:
        assert new_box.tilt is None

    # test two dimension box
    hoomd_box_2d = box.to_hoomd_convention()
    # set Lz to zero for 2D
    hoomd_box_2d[2] = 0

    lammps_box_2d = lammpsio.Box.from_hoomd_convention(hoomd_box_2d)
    # check that Lz is set to 1.0 and xz & yz are 0
    assert numpy.allclose(lammps_box_2d.high[2] - lammps_box_2d.low[2], 1.0)
    if box.tilt is not None:
        assert numpy.allclose(lammps_box_2d.tilt, [box.tilt[0], 0, 0])
    else:
        assert lammps_box_2d.tilt is None



import numpy
import packaging.version

# GSD - optional import
try:
    import gsd
    import gsd.hoomd

    # determine how GSD stores its version
    try:
        gsd_version = gsd.version.version
    except AttributeError:
        gsd_version = gsd.__version__
    gsd_version = packaging.version.Version(gsd_version)

    # GSD >= 2.8.0 deprecated Snapshot in favor of Frame
    if gsd_version >= packaging.version.Version("2.8.0"):
        gsd_frame_class = gsd.hoomd.Frame
    else:
        gsd_frame_class = gsd.hoomd.Snapshot
except ModuleNotFoundError:
    gsd_version = None

# pyzstd - optional import
try:
    import pyzstd

    pyzstd_version = packaging.version.Version(pyzstd.__version__)
except ModuleNotFoundError:
    pyzstd_version = None

# numpy
# copy behavior changed in 2.0.0, see https://github.com/scipy/scipy/pull/20172
numpy_version = packaging.version.Version(numpy.__version__)
if numpy_version >= packaging.version.Version("2.0.0"):
    numpy_copy_if_needed = None
else:
    numpy_copy_if_needed = False



from typing import Optional, Type

import numpy

from .box import Box
from .snapshot import Snapshot
from .topology import Angles, Bonds, Dihedrals, Impropers


def _readline(file_: str, require: bool = False) -> str:
    """Read and require a line."""
    line = file_.readline()
    if require and len(line) == 0:
        raise OSError("Could not read line from file")
    return line


class DataFile:
    """LAMMPS data file.

    LAMMPS can both read a data file for initialization and also write a data
    file (e.g., for visualization or restart purposes). A `Snapshot` can be
    written to a data file using `create`:

    .. code-block:: python

        snapshot = lammpsio.Snapshot(
            N=3,
            box=lammpsio.Box([-5, -5, -5], [5, 5, 5]),
            step=10
        )

        data = lammpsio.DataFile.create(tmp_path / "atoms.data", snapshot)

    A data file can also be read into a `Snapshot`:

    .. code-block:: python

        snap = data.read()

    There are many sections that can be stored in a data file, but ``lammpsio``
    does not currently understand all of them. You can check `known_headers`,
    `unknown_headers`, `known_bodies` and `unknown_bodies` for lists of what is
    currently supported.

    Parameters
    ----------
    filename : str
        Path to data file.
    atom_style : str
        LAMMPS atom style to use. Defaults to ``None``, which means the style
        should be read from the file.

    Attributes
    ----------
    filename : str
        Path to the file.
    atom_style : str
        Atom style for the file.
    known_headers : list of str
        Data file headers that can be processed.
    unknown_headers : list of str
        Data file headers that will be ignored.
    known_bodies : list of str
        Data file body sections that can be processed.
    unknown_bodies : list of str
        Data file body sections that will be ignored.

    """

    def __init__(self, filename: str, atom_style: Optional[str] = None) -> None:
        self.filename = filename
        self.atom_style = atom_style

    known_headers = (
        "atoms",
        "atom types",
        "xlo xhi",
        "ylo yhi",
        "zlo zhi",
        "xy xz yz",
        "bonds",
        "angles",
        "dihedrals",
        "impropers",
        "bond types",
        "angle types",
        "dihedral types",
        "improper types",
    )
    unknown_headers = (
        "extra bond per atom",
        "extra angle per atom",
        "extra dihedral per atom",
        "extra improper per atom",
        "extra special per atom",
        "ellipsoids",
        "lines",
        "triangles",
        "bodies",
        "avec",
        "bvec",
        "cvec",
        "abc origin",
    )
    known_bodies = (
        "Atoms",
        "Velocities",
        "Masses",
        "Bonds",
        "Angles",
        "Dihedrals",
        "Impropers",
    )
    unknown_bodies = (
        "Ellipsoids",
        "Lines",
        "Triangles",
        "Bodies",
        "Atom Type Labels",
        "Bond Type Labels",
        "Angle Type Labels",
        "Dihedral Type Labels",
        "Improper Type Labels",
        "Pair Coeffs",
        "PairIJ Coeffs",
        "Bond Coeffs",
        "Angle Coeffs",
        "Dihedral Coeffs",
        "Improper Coeffs",
        "BondBond Coeffs",
        "BondAngle Coeffs",
        "MiddleBondTorsion Coeffs",
        "EndBondTorsion Coeffs",
        "AngleTorsion Coeffs",
        "AngleAngleTorsion Coeffs",
        "BondBond13 Coeffs",
        "AngleAngle Coeffs",
    )

    @classmethod
    def create(
        cls: Type["DataFile"],
        filename: str,
        snapshot: "Snapshot",
        atom_style: Optional[str] = None,
    ) -> "DataFile":
        """Create a LAMMPS data file from a snapshot.

        Parameters
        ----------
        filename : str
            Path to data file.
        snapshot : `Snapshot`
            Snapshot to write to file.
        atom_style : str
            Atom style to use for data file. Defaults to ``None``, which means the
            style should be inferred from the contents of ``snapshot``.

        Returns
        -------
        `DataFile`
            The object representing the new data file.

        Raises
        ------
        ValueError
            If all masses are not the same for a given type.

        """

        # extract mass by type
        if snapshot.has_mass():
            masses = numpy.ones(snapshot.num_types)
            for i in range(snapshot.num_types):
                mi = snapshot.mass[snapshot.typeid == i + 1]
                if not numpy.all(mi == mi[0]):
                    raise ValueError("All masses for a type must be equal")
                elif mi[0] <= 0.0:
                    raise ValueError("Type mass must be positive value")
                masses[i] = mi[0]
        else:
            masses = None

        with open(filename, "w") as f:
            # LAMMPS header
            f.write(f"LAMMPS {filename}\n\n" f"{snapshot.N} atoms\n")

            if snapshot.has_bonds():
                f.write(f"{snapshot.bonds.N} bonds\n")

            if snapshot.has_angles():
                f.write(f"{snapshot.angles.N} angles\n")

            if snapshot.has_dihedrals():
                f.write(f"{snapshot.dihedrals.N} dihedrals\n")

            if snapshot.has_impropers():
                f.write(f"{snapshot.impropers.N} impropers\n")

            f.write(f"{snapshot.num_types} atom types\n")

            if snapshot.has_bonds():
                f.write(f"{snapshot.bonds.num_types} bond types\n")

            if snapshot.has_angles():
                f.write(f"{snapshot.angles.num_types} angle types\n")

            if snapshot.has_dihedrals():
                f.write(f"{snapshot.dihedrals.num_types} dihedral types\n")

            if snapshot.has_impropers():
                f.write(f"{snapshot.impropers.num_types} improper types\n")

            f.write(
                f"{snapshot.box.low[0]} {snapshot.box.high[0]} xlo xhi\n"
                f"{snapshot.box.low[1]} {snapshot.box.high[1]} ylo yhi\n"
                f"{snapshot.box.low[2]} {snapshot.box.high[2]} zlo zhi\n"
            )
            if snapshot.box.tilt is not None:
                f.write("{} {} {} xy xz yz\n".format(*snapshot.box.tilt))

            # Masses section
            if masses is not None:
                f.write("\nMasses\n\n")
                for i, mi in enumerate(masses):
                    f.write("{typeid:4d}{m:12}\n".format(typeid=i + 1, m=mi))

            # Atoms section
            # determine style if it is not given
            has_topology = (
                snapshot.has_bonds()
                or snapshot.has_angles()
                or snapshot.has_dihedrals()
                or snapshot.has_impropers()
            )
            if atom_style is None:
                if snapshot.has_molecule() or has_topology:
                    if snapshot.has_charge():
                        style = "full"
                    else:
                        style = "molecular"
                else:
                    if snapshot.has_charge():
                        style = "charge"
                    else:
                        style = "atomic"
            else:
                style = atom_style
            # set format string based on style
            if style == "full":
                style_fmt = (
                    "{atomid:d} {molid:d} {typeid:d} {q:.5f} {x:.8f} {y:.8f} {z:.8f}"
                )
            elif style == "charge":
                style_fmt = "{atomid:d} {typeid:d} {q:.5f} {x:.8f} {y:.8f} {z:.8f}"
            elif style == "molecular":
                style_fmt = "{atomid:d} {molid:d} {typeid:d} {x:.8f} {y:.8f} {z:.8f}"
            elif style == "atomic":
                style_fmt = "{atomid:d} {typeid:d} {x:.8f} {y:.8f} {z:.8f}"
            else:
                raise ValueError("Unknown atom style")
            if snapshot.has_image():
                style_fmt += " {ix:d} {iy:d} {iz:d}"
            # write section
            f.write(f"\nAtoms # {style}\n\n")
            for i in range(snapshot.N):
                style_args = dict(
                    atomid=snapshot.id[i] if snapshot.has_id() else i + 1,
                    typeid=snapshot.typeid[i],
                    x=snapshot.position[i][0],
                    y=snapshot.position[i][1],
                    z=snapshot.position[i][2],
                    q=snapshot.charge[i] if snapshot.has_charge() else 0.0,
                    molid=snapshot.molecule[i] if snapshot.has_molecule() else 0,
                )
                if snapshot.has_image():
                    style_args.update(
                        ix=snapshot.image[i][0],
                        iy=snapshot.image[i][1],
                        iz=snapshot.image[i][2],
                    )

                f.write(style_fmt.format(**style_args) + "\n")

            # Velocities section
            if snapshot.has_velocity():
                f.write("\nVelocities\n\n")
                for i in range(snapshot.N):
                    vel_fmt = "{atomid:8d}{vx:16.8f}{vy:16.8f}{vz:16.8f}\n"
                    f.write(
                        vel_fmt.format(
                            atomid=snapshot.id[i] if snapshot.has_id() else i + 1,
                            vx=snapshot.velocity[i][0],
                            vy=snapshot.velocity[i][1],
                            vz=snapshot.velocity[i][2],
                        )
                    )

            # Bonds section
            if snapshot.has_bonds():
                f.write("\nBonds\n\n")
                for i in range(snapshot.bonds.N):
                    f.write(
                        "{id} {typeid} {m1} {m2}\n".format(
                            id=snapshot.bonds.id[i],
                            typeid=snapshot.bonds.typeid[i],
                            m1=snapshot.bonds.members[i, 0],
                            m2=snapshot.bonds.members[i, 1],
                        )
                    )

            # Angles section
            if snapshot.has_angles():
                f.write("\nAngles\n\n")
                for i in range(snapshot.angles.N):
                    f.write(
                        "{id} {typeid} {m1} {m2} {m3}\n".format(
                            id=snapshot.angles.id[i],
                            typeid=snapshot.angles.typeid[i],
                            m1=snapshot.angles.members[i, 0],
                            m2=snapshot.angles.members[i, 1],
                            m3=snapshot.angles.members[i, 2],
                        )
                    )

            # Dihedrals section
            if snapshot.has_dihedrals():
                f.write("\nDihedrals\n\n")
                for i in range(snapshot.dihedrals.N):
                    f.write(
                        "{id} {typeid} {m1} {m2} {m3} {m4}\n".format(
                            id=snapshot.dihedrals.id[i],
                            typeid=snapshot.dihedrals.typeid[i],
                            m1=snapshot.dihedrals.members[i, 0],
                            m2=snapshot.dihedrals.members[i, 1],
                            m3=snapshot.dihedrals.members[i, 2],
                            m4=snapshot.dihedrals.members[i, 3],
                        )
                    )

            # Impropers section
            if snapshot.has_impropers():
                f.write("\nImpropers\n\n")
                for i in range(snapshot.impropers.N):
                    f.write(
                        "{id} {typeid} {m1} {m2} {m3} {m4}\n".format(
                            id=snapshot.impropers.id[i],
                            typeid=snapshot.impropers.typeid[i],
                            m1=snapshot.impropers.members[i, 0],
                            m2=snapshot.impropers.members[i, 1],
                            m3=snapshot.impropers.members[i, 2],
                            m4=snapshot.impropers.members[i, 3],
                        )
                    )
        return DataFile(filename)

    def read(self) -> "Snapshot":
        """Read a LAMMPS data file into a snapshot.

        The `atom_style` will be read from the comment in the Atoms section
        of the file. If it is not present, it must be specified manually.
        If `atom_style` is specified manually and also present in the file,
        the two must match or an error will be raised.

        Unknown headers and sections are silently ignored.

        Returns
        -------
        `Snapshot`
            Snapshot from the data file.

        """
        with open(self.filename) as f:
            # initialize snapshot from header
            N = None
            N_bonds = None
            N_angles = None
            N_dihedrals = None
            N_impropers = None
            num_types = None
            num_bond_types = None
            num_angle_types = None
            num_dihedral_types = None
            num_improper_types = None
            box_bounds = [None, None, None, None, None, None]
            box_tilt = None

            # skip first line
            _readline(f, True)
            line = _readline(f)
            while len(line) > 0:
                line = line.rstrip()

                # skip blank and go to next line
                if len(line) == 0:
                    line = _readline(f)
                    continue

                # check for unknown headers and go to next line
                skip_line = False
                for h in self.unknown_headers:
                    if h in line:
                        skip_line = True
                        break
                if skip_line:
                    line = _readline(f)
                    continue

                # line is not empty but it is not a header, so break and try to
                # make snapshot keep the line so that it can be processed in
                # next step
                done_header = True
                for h in self.known_headers:
                    if h in line:
                        done_header = False
                        break
                if done_header:
                    break

                # process useful header info
                if "atoms" in line:
                    N = int(line.split()[0])
                elif "bonds" in line:
                    N_bonds = int(line.split()[0])
                elif "angles" in line:
                    N_angles = int(line.split()[0])
                elif "dihedrals" in line:
                    N_dihedrals = int(line.split()[0])
                elif "impropers" in line:
                    N_impropers = int(line.split()[0])
                elif "atom types" in line:
                    num_types = int(line.split()[0])
                elif "bond types" in line:
                    num_bond_types = int(line.split()[0])
                elif "angle types" in line:
                    num_angle_types = int(line.split()[0])
                elif "dihedral types" in line:
                    num_dihedral_types = int(line.split()[0])
                elif "improper types" in line:
                    num_improper_types = int(line.split()[0])
                elif "xlo xhi" in line:
                    box_bounds[0], box_bounds[3] = [float(x) for x in line.split()[:2]]
                elif "ylo yhi" in line:
                    box_bounds[1], box_bounds[4] = [float(x) for x in line.split()[:2]]
                elif "zlo zhi" in line:
                    box_bounds[2], box_bounds[5] = [float(x) for x in line.split()[:2]]
                elif "xy xz yz" in line:
                    box_tilt = [float(x) for x in line.split()[:3]]
                else:
                    raise RuntimeError("Uncaught header line! Check programming")

                # done here, read next line
                line = _readline(f)

            if N is None:
                raise IOError("Number of particles not read")
            elif num_types is None:
                raise IOError("Number of types not read")
            elif None in box_bounds:
                raise IOError("Box bounds not read")
            box = Box(box_bounds[:3], box_bounds[3:], box_tilt)
            snap = Snapshot(N, box, num_types=num_types)
            id_map = {}

            # now that snapshot is made, file it in with body sections
            masses = None
            while len(line) > 0:
                if "Atoms" in line:
                    # use or extract style
                    row = line.split()
                    if len(row) == 3:
                        style = row[-1]
                        if self.atom_style is not None and style != self.atom_style:
                            raise ValueError(
                                "Specified style does not match style in file"
                            )
                    else:
                        style = self.atom_style
                    if style is None:
                        raise IOError("Atom style not found, specify.")
                    # number of columns to read for style
                    if style == "full":
                        style_cols = 3
                    elif style == "charge":
                        style_cols = 2
                    elif style == "molecular":
                        style_cols = 2
                    elif style == "atomic":
                        style_cols = 1
                    else:
                        raise ValueError("Unknown atom style")

                    # read atom coordinates
                    _readline(f, True)  # blank line
                    for i in range(snap.N):
                        # strip out comments
                        row = _readline(f, True).split()
                        try:
                            comment = row.index("#")
                            row = row[:comment]
                        except ValueError:
                            pass

                        # check that row is correctly sized and process
                        if not (
                            len(row) == style_cols + 4 or len(row) == style_cols + 7
                        ):
                            raise IOError(
                                "Expected number of columns not read for atom style"
                            )

                        # only save the atom id if it is not in standard order
                        id_ = int(row[0])
                        if id_ != i + 1:
                            if id_ not in id_map:
                                id_map[id_] = i
                            idx = id_map[id_]
                            snap.id[idx] = id_
                        else:
                            idx = i

                        if style == "full":
                            snap.molecule[idx] = int(row[1])
                            snap.typeid[idx] = int(row[2])
                            snap.charge[idx] = float(row[3])
                        elif style == "charge":
                            snap.typeid[idx] = int(row[1])
                            snap.charge[idx] = float(row[2])
                        elif style == "molecular":
                            snap.molecule[idx] = int(row[1])
                            snap.typeid[idx] = int(row[2])
                        elif style == "atomic":
                            snap.typeid[idx] = int(row[1])
                        snap.position[idx] = [
                            float(x) for x in row[style_cols + 1 : style_cols + 4]
                        ]
                        if len(row) == style_cols + 7:
                            snap.image[idx] = [
                                int(x) for x in row[style_cols + 4 : style_cols + 7]
                            ]

                    # sanity check types
                    if numpy.any(
                        numpy.logical_or(snap.typeid < 1, snap.typeid > num_types)
                    ):
                        raise ValueError("Invalid type id")
                elif "Velocities" in line:
                    _readline(f, True)  # blank line
                    for i in range(snap.N):
                        row = _readline(f, True).split()
                        if len(row) < 4:
                            raise IOError(
                                "Expected number of columns not read for velocity"
                            )
                        # parse atom id: need to repeat mapping in case
                        # Velocity comes before Atoms
                        id_ = int(row[0])
                        if id_ != i + 1:
                            if id_ not in id_map:
                                id_map[id_] = i
                            idx = id_map[id_]
                            snap.id[idx] = id_
                        else:
                            idx = i
                        snap.velocity[idx] = [float(x) for x in row[1:4]]
                elif "Masses" in line:
                    masses = {}
                    _readline(f, True)  # blank line
                    for i in range(num_types):
                        row = _readline(f, True).split()
                        if len(row) < 2:
                            raise IOError(
                                "Expected number of columns not read for mass"
                            )
                        masses[int(row[0])] = float(row[1])
                elif "Bonds" in line:
                    if N_bonds is not None:
                        snap.bonds = Bonds(N_bonds, num_bond_types)
                    _readline(f, True)  # blank line
                    for i in range(snap.bonds.N):
                        row = _readline(f, True).split()
                        if len(row) < 4:
                            raise IOError(
                                "Expected number of columns not read for bonds"
                            )
                        row = [int(x) for x in row]
                        # only write ID if it is out of default order
                        id = row[0]
                        if id != i + 1:
                            snap.bonds.id[i] = id
                        snap.bonds.typeid[i] = row[1]
                        snap.bonds.members[i] = row[2:4]

                    # sanity check
                    if numpy.any(snap.bonds.num_types < 1) or numpy.any(
                        snap.bonds.typeid > snap.bonds.num_types
                    ):
                        raise ValueError("Invalid bond type id")
                elif "Angles" in line:
                    if N_angles is not None:
                        snap.angles = Angles(N_angles, num_angle_types)
                    _readline(f, True)  # blank line
                    for i in range(snap.angles.N):
                        row = _readline(f, True).split()
                        if len(row) < 5:
                            raise IOError(
                                "Expected number of columns not read for angles"
                            )
                        row = [int(x) for x in row]
                        # only write ID if it is out of default order
                        id = row[0]
                        if id != i + 1:
                            snap.angles.id[i] = id
                        snap.angles.typeid[i] = row[1]
                        snap.angles.members[i] = row[2:5]

                    # sanity check
                    if numpy.any(snap.angles.num_types < 1) or numpy.any(
                        snap.angles.typeid > snap.angles.num_types
                    ):
                        raise ValueError("Invalid angle type id")
                elif "Dihedrals" in line:
                    if N_dihedrals is not None:
                        snap.dihedrals = Dihedrals(N_dihedrals, num_dihedral_types)
                    _readline(f, True)  # blank line
                    for i in range(snap.dihedrals.N):
                        row = _readline(f, True).split()
                        if len(row) < 6:
                            raise IOError(
                                "Expected number of columns not read for dihedrals"
                            )
                        row = [int(x) for x in row]
                        # only write ID if it is out of default order
                        id = row[0]
                        if id != i + 1:
                            snap.dihedrals.id[i] = id
                        snap.dihedrals.typeid[i] = row[1]
                        snap.dihedrals.members[i] = row[2:6]

                    # sanity check
                    if numpy.any(snap.dihedrals.num_types < 1) or numpy.any(
                        snap.dihedrals.typeid > snap.dihedrals.num_types
                    ):
                        raise ValueError("Invalid dihedral type id")
                elif "Impropers" in line:
                    if N_impropers is not None:
                        snap.impropers = Impropers(N_impropers, num_improper_types)
                    _readline(f, True)  # blank line
                    for i in range(snap.impropers.N):
                        row = _readline(f, True).split()
                        if len(row) < 6:
                            raise IOError(
                                "Expected number of columns not read for impropers"
                            )
                        row = [int(x) for x in row]
                        # only write ID if it is out of default order
                        id = row[0]
                        if id != i + 1:
                            snap.impropers.id[i] = id
                        snap.impropers.typeid[i] = row[1]
                        snap.impropers.members[i] = row[2:6]

                    # sanity check
                    if numpy.any(snap.impropers.num_types < 1) or numpy.any(
                        snap.impropers.typeid > snap.impropers.num_types
                    ):
                        raise ValueError("Invalid improper type id")
                else:
                    # silently ignore unknown sections / lines
                    pass

                line = _readline(f)

            # set mass on particles at end, in case sections were out of order in file
            if masses is not None:
                for typeid, m in masses.items():
                    snap.mass[snap.typeid == typeid] = m

        return snap



import copy

import numpy
import pytest
from packaging import version

import lammpsio

try:
    import gsd.hoomd

    has_gsd = True
except ModuleNotFoundError:
    has_gsd = False


def test_create(snap):
    assert snap.N == 3
    assert isinstance(snap.box, lammpsio.Box)
    assert snap.step == 10


@pytest.mark.skipif(not has_gsd, reason="gsd not installed")
def test_gsd_conversion():
    # make a GSD frame
    try:
        gsd_version = gsd.version.version
    except AttributeError:
        gsd_version = gsd.__version__
    if version.Version(gsd_version) >= version.Version("2.8.0"):
        frame = gsd.hoomd.Frame()
    else:
        frame = gsd.hoomd.Snapshot()
    frame.configuration.step = 3
    frame.configuration.box = [4, 5, 6, 0.1, 0.2, 0.3]
    frame.particles.N = 2
    frame.particles.position = [[0.1, 0.2, 0.3], [-0.1, -0.2, -0.3]]
    frame.particles.image = [[1, -1, 0], [0, 2, -2]]
    frame.particles.velocity = [[1, 2, 3], [-4, -5, -6]]
    frame.particles.body = [0, -1]
    frame.particles.types = ["A", "B"]
    frame.particles.typeid = [1, 0]
    frame.particles.mass = [3, 2]
    frame.particles.charge = [-1, 1]

    # make Snapshot from GSD
    snap, type_map = lammpsio.Snapshot.from_hoomd_gsd(frame)
    assert snap.step == 3
    assert numpy.allclose(snap.box.low, [-2.85, -3.4, -3])
    assert numpy.allclose(snap.box.high, [1.15, 1.6, 3])
    assert numpy.allclose(snap.box.tilt, [0.5, 1.2, 1.8])
    assert snap.N == 2
    assert numpy.allclose(snap.position, [[0.1, 0.2, 0.3], [-0.1, -0.2, -0.3]])
    assert numpy.all(snap.image == [[1, -1, 0], [0, 2, -2]])
    assert numpy.allclose(snap.velocity, [[1, 2, 3], [-4, -5, -6]])
    assert numpy.all(snap.molecule == [1, 0])
    assert numpy.all(snap.typeid == [2, 1])
    assert snap.type_label == {1: "A", 2: "B"}
    assert numpy.allclose(snap.mass, [3, 2])
    assert numpy.allclose(snap.charge, [-1, 1])
    assert type_map == {1: "A", 2: "B"}

    # go back to GSD frame
    frame2 = snap.to_hoomd_gsd()
    assert frame2.configuration.step == frame.configuration.step
    assert numpy.allclose(frame2.configuration.box, frame.configuration.box)
    assert frame2.particles.N == frame.particles.N
    assert numpy.allclose(frame2.particles.position, frame.particles.position)
    assert numpy.all(frame2.particles.image == frame.particles.image)
    assert numpy.allclose(frame2.particles.velocity, frame.particles.velocity)
    assert numpy.all(frame2.particles.types == tuple(frame.particles.types))
    assert numpy.all(frame2.particles.typeid == frame.particles.typeid)
    assert numpy.allclose(frame2.particles.mass, frame.particles.mass)
    assert numpy.allclose(frame2.particles.charge, frame.particles.charge)
    assert numpy.all(frame2.particles.body == frame.particles.body)

    # do the same thing, but use explicit type map (should give warning)
    with pytest.warns(DeprecationWarning):
        frame3 = snap.to_hoomd_gsd({1: "C", 2: "D"})
    assert numpy.all(frame3.particles.types == ("C", "D"))
    assert numpy.all(frame3.particles.typeid == [1, 0])

    # do the same thing, but with different type_label
    snap.type_label = lammpsio.topology.LabelMap({1: "C", 2: "D"})
    frame4 = snap.to_hoomd_gsd()
    assert numpy.all(frame4.particles.types == tuple(["C", "D"]))
    assert numpy.all(frame4.particles.typeid == [1, 0])

    # check for warning on floppy molecules
    frame.particles.body = [-2, -1]
    with pytest.warns():
        lammpsio.Snapshot.from_hoomd_gsd(frame)

    # check for barebones Snapshot going to GSD
    snap2 = lammpsio.Snapshot(N=2, box=lammpsio.Box([-5, -5, -5], [5, 5, 5]))
    snap2.to_hoomd_gsd()
    # check again with id remapping to make sure order is preserved
    snap2.id = [2, 1]
    snap2.to_hoomd_gsd()
    assert numpy.all(snap2.id == [2, 1])

    # check that orthorhombic box is set correctly when not originally centered
    snap3 = lammpsio.Snapshot(N=2, box=lammpsio.Box([0, 0, 0], [15, 15, 15]))
    snap3.position = [[0, 0, 0], [1, 1, 1]]
    # center calculated using box vectors as describe in LAMMPS documentation
    center = [7.5, 7.5, 7.5]
    frame5 = snap3.to_hoomd_gsd()
    assert numpy.allclose(frame5.configuration.box, [15, 15, 15, 0, 0, 0])
    assert numpy.allclose(frame5.particles.position, snap3.position - center)

    # check that triclinic box is set correctly when not originally centered
    snap3.box.tilt = [1.5, 3.0, 4.5]
    # center calculated using box vectors as describe in LAMMPS documentation
    center = [9.75, 9.75, 7.5]
    frame6 = snap3.to_hoomd_gsd()
    assert numpy.allclose(frame6.configuration.box, [15, 15, 15, 0.1, 0.2, 0.3])
    assert numpy.allclose(frame6.particles.position, snap3.position - center)


@pytest.mark.skipif(not has_gsd, reason="gsd not installed")
def test_minimal_gsd_conversion():
    # make a GSD frame
    try:
        gsd_version = gsd.version.version
    except AttributeError:
        gsd_version = gsd.__version__
    if version.Version(gsd_version) >= version.Version("2.8.0"):
        frame = gsd.hoomd.Frame()
    else:
        frame = gsd.hoomd.Snapshot()

    frame.configuration.box = [4, 5, 6, 0.1, 0.2, 0.3]
    frame.particles.N = 2
    frame.particles.types = ["A", "B"]
    frame.particles.typeid = [1, 0]
    frame.particles.position = [[0.1, 0.2, 0.3], [-0.1, -0.2, -0.3]]

    # make Snapshot from GSD with minimial information
    snap, type_map = lammpsio.Snapshot.from_hoomd_gsd(frame)
    # check that unspecified properties are not set
    assert snap.has_velocity() is False
    assert snap.has_image() is False
    assert snap.has_charge() is False
    assert snap.has_mass() is False
    assert snap.has_molecule() is False
    assert snap.has_bonds() is False
    assert snap.has_angles() is False
    assert snap.has_dihedrals() is False
    assert snap.has_impropers() is False


@pytest.mark.skipif(not has_gsd, reason="gsd not installed")
def test_gsd_conversion_topology():
    # make a GSD frame
    try:
        gsd_version = gsd.version.version
    except AttributeError:
        gsd_version = gsd.__version__
    if version.Version(gsd_version) >= version.Version("2.8.0"):
        frame = gsd.hoomd.Frame()
    else:
        frame = gsd.hoomd.Snapshot()

    # set box
    frame.configuration.box = [10, 10, 10, 1, 1, 1]

    # particles
    position = [
        [0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [1.0, 0.0, 0.0],
        [1.0, 1.0, 0.0],
        [3.0, 3.0, 0.0],
        [3.0, 4.0, 0.0],
        [4.0, 3.0, 0.0],
        [4.0, 4.0, 0.0],
    ]
    frame.particles.N = 8
    frame.particles.types = ["A", "B"]
    frame.particles.typeid = [0, 0, 0, 0, 1, 1, 1, 1]
    frame.particles.position = position

    # bonds
    frame.bonds.N = 6
    frame.bonds.types = ["bondA", "bondB"]
    frame.bonds.typeid = [0, 1, 0, 1, 0, 1]
    frame.bonds.group = [
        [0, 1],
        [1, 2],
        [2, 3],
        [4, 5],
        [5, 6],
        [6, 7],
    ]

    # angles
    frame.angles.N = 4
    frame.angles.types = ["angleA", "angleB"]
    frame.angles.typeid = [0, 1, 0, 1]
    frame.angles.group = [
        [0, 1, 2],
        [1, 2, 3],
        [4, 5, 6],
        [5, 6, 7],
    ]

    # dihedrals
    frame.dihedrals.N = 2
    frame.dihedrals.types = ["dihedralA", "dihedralB"]
    frame.dihedrals.typeid = [0, 1]
    frame.dihedrals.group = [
        [0, 1, 2, 3],
        [4, 5, 6, 7],
    ]

    # impropers
    frame.impropers.N = 2
    frame.impropers.types = ["improperA", "improperB"]
    frame.impropers.typeid = [0, 1]
    frame.impropers.group = [
        [0, 1, 2, 3],
        [4, 5, 6, 7],
    ]

    # make Snapshot from GSD
    snap, type_map = lammpsio.Snapshot.from_hoomd_gsd(frame)
    # particles
    assert snap.N == 8
    assert numpy.allclose(snap.position, position)
    assert numpy.all(snap.typeid == [1, 1, 1, 1, 2, 2, 2, 2])
    assert type_map == {1: "A", 2: "B"}
    assert snap.type_label == {1: "A", 2: "B"}
    # bonds
    assert snap.bonds.N == 6
    assert numpy.all(snap.bonds.typeid == [1, 2, 1, 2, 1, 2])
    assert numpy.all(
        snap.bonds.members == [[1, 2], [2, 3], [3, 4], [5, 6], [6, 7], [7, 8]]
    )
    assert snap.bonds.type_label == {1: "bondA", 2: "bondB"}
    # angles
    assert snap.angles.N == 4
    assert numpy.all(snap.angles.typeid == [1, 2, 1, 2])
    assert numpy.all(
        snap.angles.members == [[1, 2, 3], [2, 3, 4], [5, 6, 7], [6, 7, 8]]
    )
    assert snap.angles.type_label == {1: "angleA", 2: "angleB"}
    # dihedrals
    assert snap.dihedrals.N == 2
    assert numpy.all(snap.dihedrals.typeid == [1, 2])
    assert numpy.all(snap.dihedrals.members == [[1, 2, 3, 4], [5, 6, 7, 8]])
    assert snap.dihedrals.type_label == {1: "dihedralA", 2: "dihedralB"}
    # impropers
    assert snap.impropers.N == 2
    assert numpy.all(snap.impropers.typeid == [1, 2])
    assert numpy.all(snap.impropers.members == [[1, 2, 3, 4], [5, 6, 7, 8]])
    assert snap.impropers.type_label == {1: "improperA", 2: "improperB"}

    # go back to GSD frame
    frame2 = snap.to_hoomd_gsd()
    assert numpy.allclose(frame2.configuration.box, frame.configuration.box)
    # particles
    assert frame2.particles.N == frame.particles.N
    assert numpy.allclose(frame2.particles.position, frame.particles.position)
    assert numpy.all(frame2.particles.types == tuple(frame.particles.types))
    assert numpy.all(frame2.particles.typeid == frame.particles.typeid)
    # bonds
    assert frame2.bonds.N == frame.bonds.N
    assert numpy.all(frame2.bonds.types == tuple(frame.bonds.types))
    assert numpy.all(frame2.bonds.typeid == frame.bonds.typeid)
    assert numpy.all(frame2.bonds.group == frame.bonds.group)
    # angles
    assert frame2.angles.N == frame.angles.N
    assert numpy.all(frame2.angles.types == tuple(frame.angles.types))
    assert numpy.all(frame2.angles.typeid == frame.angles.typeid)
    assert numpy.all(frame2.angles.group == frame.angles.group)
    # dihedrals
    assert frame2.dihedrals.N == frame.dihedrals.N
    assert numpy.all(frame2.dihedrals.types == tuple(frame.dihedrals.types))
    assert numpy.all(frame2.dihedrals.typeid == frame.dihedrals.typeid)
    assert numpy.all(frame2.dihedrals.group == frame.dihedrals.group)
    # impropers
    assert frame2.impropers.N == frame.impropers.N
    assert numpy.all(frame2.impropers.types == tuple(frame.impropers.types))
    assert numpy.all(frame2.impropers.typeid == frame.impropers.typeid)
    assert numpy.all(frame2.impropers.group == frame.impropers.group)


def test_position(snap):
    assert not snap.has_position()
    assert numpy.allclose(snap.position, 0.0)
    position = [[0.1, 0.2, 0.3], [-0.4, -0.5, -0.6], [100, -200, 300]]
    snap.position = position
    assert numpy.allclose(snap.position, position)
    with pytest.raises(TypeError):
        snap.position = [[0, 0, 0], [0, 0, 0]]
    with pytest.raises(TypeError):
        snap.position = [0, 0, 0]


def test_image(snap):
    assert not snap.has_image()
    assert numpy.allclose(snap.image, 0)
    image = [[1, 2, 3], [-1, 0, -2], [100, -200, 300]]
    snap.image = image
    assert numpy.allclose(snap.image, image)
    with pytest.raises(TypeError):
        snap.image = [[0, 0, 0], [0, 0, 0]]
    with pytest.raises(TypeError):
        snap.image = [0, 0, 0]


def test_velocity(snap):
    assert not snap.has_velocity()
    assert numpy.allclose(snap.velocity, 0.0)
    velocity = [[0.1, 0.2, 0.3], [-0.4, -0.5, -0.6], [100, -200, 300]]
    snap.velocity = velocity
    assert numpy.allclose(snap.velocity, velocity)
    with pytest.raises(TypeError):
        snap.velocity = [[0, 0, 0], [0, 0, 0]]
    with pytest.raises(TypeError):
        snap.velocity = [0, 0, 0]


def test_typeid(snap):
    assert not snap.has_typeid()
    assert numpy.allclose(snap.typeid, 1)
    typeid = [2, 1, 3]
    snap.typeid = typeid
    assert numpy.allclose(snap.typeid, typeid)
    with pytest.raises(TypeError):
        snap.typeid = [1, 1]


def test_molecule(snap):
    assert not snap.has_molecule()
    assert numpy.allclose(snap.molecule, 0)
    molecule = [2, 0, 1]
    snap.molecule = molecule
    assert numpy.allclose(snap.molecule, molecule)
    with pytest.raises(TypeError):
        snap.molecule = [0, 0]


def test_mass(snap):
    assert not snap.has_mass()
    assert numpy.allclose(snap.mass, 1)
    # also set typeid, so that types and masses are consistent as LAMMPS requires
    # removing this line doesn't actually do anything, though
    snap.typeid = [1, 2, 1]
    mass = [2, 3, 2]
    snap.mass = mass
    assert numpy.allclose(snap.mass, mass)
    with pytest.raises(TypeError):
        snap.mass = [1, 1]


def test_bonds(snap_8):
    snap_8.bonds = lammpsio.topology.Bonds(N=6, num_types=2)
    id = [1, 2, 3, 4, 5, 6]
    typeid = [1, 2, 1, 2, 1, 2]
    members = [
        [1, 2],
        [2, 3],
        [3, 4],
        [5, 6],
        [6, 7],
        [7, 8],
    ]

    snap_8.bonds.id = id
    snap_8.bonds.typeid = typeid
    snap_8.bonds.members = members

    assert numpy.allclose(snap_8.bonds.id, id)
    assert numpy.allclose(snap_8.bonds.typeid, typeid)
    assert numpy.allclose(snap_8.bonds.members, members)


def test_angles(snap_8):
    snap_8.angles = lammpsio.topology.Angles(N=4, num_types=2)
    id = [
        1,
        2,
        3,
        4,
    ]
    typeid = [1, 2, 1, 2]
    members = [
        [1, 2, 3],
        [2, 3, 4],
        [5, 6, 7],
        [6, 7, 8],
    ]

    snap_8.angles.id = id
    snap_8.angles.typeid = typeid
    snap_8.angles.members = members

    assert numpy.allclose(snap_8.angles.id, id)
    assert numpy.allclose(snap_8.angles.typeid, typeid)
    assert numpy.allclose(snap_8.angles.members, members)


def test_dihedrals(snap_8):
    snap_8.dihedrals = lammpsio.topology.Dihedrals(N=2, num_types=2)
    id = [1, 2]
    typeid = [1, 2]
    members = [
        [1, 2, 3, 4],
        [5, 6, 7, 8],
    ]

    snap_8.dihedrals.id = id
    snap_8.dihedrals.typeid = typeid
    snap_8.dihedrals.members = members

    assert numpy.allclose(snap_8.dihedrals.id, id)
    assert numpy.allclose(snap_8.dihedrals.typeid, typeid)
    assert numpy.allclose(snap_8.dihedrals.members, members)


def test_impropers(snap_8):
    snap_8.impropers = lammpsio.topology.Impropers(N=2, num_types=2)
    id = [1, 2]
    typeid = [1, 2]
    members = [
        [1, 2, 3, 4],
        [5, 6, 7, 8],
    ]

    snap_8.impropers.id = id
    snap_8.impropers.typeid = typeid
    snap_8.impropers.members = members

    assert numpy.allclose(snap_8.impropers.id, id)
    assert numpy.allclose(snap_8.impropers.typeid, typeid)
    assert numpy.allclose(snap_8.impropers.members, members)


def test_charge(snap):
    assert not snap.has_charge()
    assert numpy.allclose(snap.charge, 0)
    charge = [-1, 0, 1]
    snap.charge = charge
    assert numpy.allclose(snap.charge, charge)
    with pytest.raises(TypeError):
        snap.typeid = [0, 0]


def test_copy(snap):
    snap.id = [2, 0, 1]
    snap.position = [[0.1, 0.2, 0.3], [-0.4, -0.5, -0.6], [0.7, 0.8, 0.9]]
    snap.image = [[1, 2, 3], [-4, -5, -6], [7, 8, 9]]
    snap.velocity = [[-3, -2, -1], [6, 5, 4], [9, 8, 7]]
    snap.typeid = [2, 1, 2]
    snap.mass = [3, 2, 3]
    snap.molecule = [2, 0, 1]
    snap.charge = [-1, 0, 1]

    # check shallow copy works
    snap_shallow = copy.copy(snap)
    assert numpy.all(snap_shallow.id == snap.id)
    assert numpy.allclose(snap_shallow.position, snap.position)
    assert numpy.all(snap_shallow.image == snap.image)
    assert numpy.allclose(snap_shallow.velocity, snap.velocity)
    assert numpy.all(snap_shallow.typeid == snap.typeid)
    assert numpy.allclose(snap_shallow.mass, snap.mass)
    assert numpy.all(snap_shallow.molecule == snap.molecule)
    assert numpy.allclose(snap_shallow.charge, snap.charge)

    # check deep copy works
    snap_deep = copy.deepcopy(snap)
    assert numpy.all(snap_deep.id == snap.id)
    assert numpy.allclose(snap_deep.position, snap.position)
    assert numpy.all(snap_deep.image == snap.image)
    assert numpy.allclose(snap_deep.velocity, snap.velocity)
    assert numpy.all(snap_deep.typeid == snap.typeid)
    assert numpy.allclose(snap_deep.mass, snap.mass)
    assert numpy.all(snap_deep.molecule == snap.molecule)
    assert numpy.allclose(snap_deep.charge, snap.charge)

    # change the original snapshot and check shallow changes, but deep does not
    old_ids = list(snap.id)
    snap.id = [3, 4, 5]
    assert numpy.all(snap_shallow.id == snap.id)
    assert numpy.all(snap_deep.id == old_ids)



import pytest

import lammpsio


def test_bonds(snap_8):
    # default is no bonds
    assert snap_8.bonds is None
    assert not snap_8.has_bonds()

    # empty bonds set still means we don't have bonds
    bonds = lammpsio.Bonds(N=0)
    snap_8.bonds = bonds
    assert snap_8.bonds is bonds
    assert not snap_8.has_bonds()

    # one bond counts as bonds
    bonds = lammpsio.Bonds(N=1)
    snap_8.bonds = bonds
    assert snap_8.bonds is bonds
    assert snap_8.has_bonds()

    # make sure can set back to None
    snap_8.bonds = None
    assert snap_8.bonds is None
    assert not snap_8.has_bonds()

    # make sure other types cannot be set and nothing changes
    with pytest.raises(TypeError):
        snap_8.bonds = lammpsio.Angles(N=0)
    assert snap_8.bonds is None
    assert not snap_8.has_bonds()


def test_angles(snap_8):
    # default is no angles
    assert snap_8.angles is None
    assert not snap_8.has_angles()

    # empty angles set still means we don't have angles
    angles = lammpsio.Angles(N=0)
    snap_8.angles = angles
    assert snap_8.angles is angles
    assert not snap_8.has_angles()

    # one bond counts as angles
    angles = lammpsio.Angles(N=1)
    snap_8.angles = angles
    assert snap_8.angles is angles
    assert snap_8.has_angles()

    # make sure can set back to None
    snap_8.angles = None
    assert snap_8.angles is None
    assert not snap_8.has_angles()

    # make sure other types cannot be set and nothing changes
    with pytest.raises(TypeError):
        snap_8.angles = lammpsio.Bonds(N=0)
    assert snap_8.angles is None
    assert not snap_8.has_angles()


def test_dihedrals(snap_8):
    # default is no dihedrals
    assert snap_8.dihedrals is None
    assert not snap_8.has_dihedrals()

    # empty dihedrals set still means we don't have dihedrals
    dihedrals = lammpsio.Dihedrals(N=0)
    snap_8.dihedrals = dihedrals
    assert snap_8.dihedrals is dihedrals
    assert not snap_8.has_dihedrals()

    # one bond counts as dihedrals
    dihedrals = lammpsio.Dihedrals(N=1)
    snap_8.dihedrals = dihedrals
    assert snap_8.dihedrals is dihedrals
    assert snap_8.has_dihedrals()

    # make sure can set back to None
    snap_8.dihedrals = None
    assert snap_8.dihedrals is None
    assert not snap_8.has_dihedrals()

    # make sure other types cannot be set and nothing changes
    with pytest.raises(TypeError):
        snap_8.dihedrals = lammpsio.Angles(N=0)
    assert snap_8.dihedrals is None
    assert not snap_8.has_dihedrals()


def test_impropers(snap_8):
    # default is no impropers
    assert snap_8.impropers is None
    assert not snap_8.has_impropers()

    # empty impropers set still means we don't have impropers
    impropers = lammpsio.Impropers(N=0)
    snap_8.impropers = impropers
    assert snap_8.impropers is impropers
    assert not snap_8.has_impropers()

    # one bond counts as impropers
    impropers = lammpsio.Impropers(N=1)
    snap_8.impropers = impropers
    assert snap_8.impropers is impropers
    assert snap_8.has_impropers()

    # make sure can set back to None
    snap_8.impropers = None
    assert snap_8.impropers is None
    assert not snap_8.has_impropers()

    # make sure other types cannot be set and nothing changes
    with pytest.raises(TypeError):
        snap_8.impropers = lammpsio.Angles(N=0)
    assert snap_8.impropers is None
    assert not snap_8.has_impropers()


def test_bonds_wrong_shape(snap_8):
    snap_8.bonds = lammpsio.topology.Bonds(N=6, num_types=2)

    # check that error is raised if array is the wrong shape
    id = [1, 2, 3, 4, 5, 6, 7]
    with pytest.raises(TypeError):
        snap_8.bonds.id = id

    typeid = [1, 1, 1, 2, 2, 2, 3]
    with pytest.raises(TypeError):
        snap_8.bonds.typeid = typeid

    members = [[1, 2], [2, 3], [3, 4], [5, 6], [6, 7], [7, 8], [8, 9]]
    with pytest.raises(TypeError):
        snap_8.bonds.members = members


def test_angles_wrong_shape(snap_8):
    snap_8.angles = lammpsio.topology.Angles(N=4, num_types=2)

    # check that error is raised if array is the wrong shape
    id = [1, 2, 3, 4, 5]
    with pytest.raises(TypeError):
        snap_8.angles.id = id

    typeid = [1, 1, 2, 2, 3]
    with pytest.raises(TypeError):
        snap_8.angles.typeid = typeid

    members = [
        [1, 2, 3],
        [2, 3, 4],
        [5, 6, 7],
        [6, 7, 8],
        [7, 8, 9],
    ]
    with pytest.raises(TypeError):
        snap_8.angles.members = members


def test_dihedrals_wrong_shape(snap_8):
    snap_8.dihedrals = lammpsio.topology.Dihedrals(N=2, num_types=2)

    # check that error is raised if array is the wrong shape
    id = [1, 2, 3]
    with pytest.raises(TypeError):
        snap_8.dihedrals.id = id

    typeid = [1, 2, 3]
    with pytest.raises(TypeError):
        snap_8.dihedrals.typeid = typeid

    members = [
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        [6, 7, 8, 9],
    ]
    with pytest.raises(TypeError):
        snap_8.dihedrals.members = members


def test_impropers_wrong_shape(snap_8):
    snap_8.impropers = lammpsio.topology.Impropers(N=2, num_types=2)

    # check that error is raised if array is the wrong shape
    id = [1, 2, 3]
    with pytest.raises(TypeError):
        snap_8.impropers.id = id

    typeid = [1, 2, 3]
    with pytest.raises(TypeError):
        snap_8.impropers.typeid = typeid

    members = [
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        [6, 7, 8, 9],
    ]
    with pytest.raises(TypeError):
        snap_8.impropers.members = members


def test_LabelMap():
    # create a simple label map
    label = lammpsio.topology.LabelMap({1: "typeA", 2: "typeB"})
    # check the types
    assert label.types == ("typeA", "typeB")
    # check the typeids
    assert label.typeid == (1, 2)
    # get
    assert label[2] == "typeB"
    # get inverse
    assert label.inverse["typeB"] == 2
    # set
    label[3] = "typeC"
    assert label[3] == "typeC"
    assert label.inverse["typeC"] == 3
    # delete
    del label[3]
    assert 3 not in label
    assert "typeC" not in label.inverse



"""Topology (connection and type) information."""

import collections.abc
from typing import Any, Optional

import numpy

from . import _compatibility


class Topology:
    """Particle topology.

    Parameters
    ----------
    N : int
        Number of connections.
    num_members : int
        Number of members in a connection.
    num_types : int
        Number of connection types. Default of ``None`` means the number of
        types is determined from the unique typeids.

    All values of indexes follow the LAMMPS 1-indexed convention, but the
    arrays themselves are 0-indexed. Lazy array initialization is used as for
    the `Snapshot`.

    """

    def __init__(
        self, N: int, num_members: int, num_types: Optional[int] = None
    ) -> None:
        self._N = N
        self._num_members = num_members
        self.num_types = num_types

        self._type_label = None
        self._id = None
        self._typeid = None
        self._members = None

    @property
    def N(self) -> int:
        """int: Number of connections."""
        return self._N

    @property
    def id(self) -> numpy.ndarray:
        """(*N*,) `numpy.ndarray` of `int`: Unique identifiers (IDs).

        The default value on initialization runs from 1 to `N`.

        """
        if not self.has_id():
            self._id = numpy.arange(1, self.N + 1)
        return self._id

    @id.setter
    def id(self, value: Any) -> None:
        if value is not None:
            v = numpy.array(
                value, ndmin=1, copy=_compatibility.numpy_copy_if_needed, dtype=int
            )
            if v.shape != (self.N,):
                raise TypeError("Ids must be a size N array")
            if not self.has_id():
                self._id = numpy.arange(1, self.N + 1)
            numpy.copyto(self._id, v)
        else:
            self._id = None

    def has_id(self) -> bool:
        """Check if configuration has connection IDs.

        Returns
        -------
        bool
            ``True`` if connection IDs have been initialized.

        """
        return self._id is not None

    @property
    def typeid(self) -> numpy.ndarray:
        """(*N*,) `numpy.ndarray` of `int`: Connection type IDs.

        The default value on initialization is 1 for all entries.

        """
        if not self.has_typeid():
            self._typeid = numpy.ones(self.N, dtype=int)
        return self._typeid

    @typeid.setter
    def typeid(self, value: Any) -> None:
        if value is not None:
            v = numpy.array(
                value, ndmin=1, copy=_compatibility.numpy_copy_if_needed, dtype=int
            )
            if v.shape != (self.N,):
                raise TypeError("typeids must be a size N array")
            if not self.has_typeid():
                self._typeid = numpy.ones(self.N, dtype=int)
            numpy.copyto(self._typeid, v)
        else:
            self._typeid = None

    def has_typeid(self) -> bool:
        """Check if configuration has connection typeids.

        Returns
        -------
        bool
            ``True`` if connection typeids have been initialized.

        """
        return self._typeid is not None

    @property
    def members(self) -> numpy.ndarray:
        """(*N*, *M*) `numpy.ndarray` of `int`: Connection members.

        The default value on initialization is 1 for all entries. *M* is the
        number of members in the connection.

        """
        if not self.has_members():
            self._members = numpy.ones((self.N, self._num_members), dtype=int)
        return self._members

    @members.setter
    def members(self, value: Any) -> None:
        if value is not None:
            v = numpy.array(
                value, ndmin=2, copy=_compatibility.numpy_copy_if_needed, dtype=int
            )
            if v.shape != (self.N, self._num_members):
                raise TypeError("Members must be a size N x number of members array")
            if not self.has_members():
                self._members = numpy.ones((self.N, self._num_members), dtype=int)
            numpy.copyto(self._members, v)
        else:
            self._members = None

    def has_members(self) -> bool:
        """Check if configuration has connection members.

        Returns
        -------
        bool
            ``True`` if particle members have been initialized.

        """
        return self._members is not None

    @property
    def num_types(self) -> int:
        """int: Number of connection types."""
        if self._num_types is not None:
            return self._num_types
        else:
            if self.has_typeid():
                return numpy.amax(self.typeid)
            else:
                return 1

    @num_types.setter
    def num_types(self, value: Any) -> None:
        if value is not None:
            self._num_types = int(value)
        else:
            self._num_types = None

    @property
    def type_label(self) -> Optional["LabelMap"]:
        """LabelMap: Labels of connection type IDs."""

        return self._type_label

    @type_label.setter
    def type_label(self, value: Optional["LabelMap"]) -> None:
        if value is not None:
            if not isinstance(value, LabelMap):
                raise TypeError("type_label must be a LabelMap")
            self._type_label = value
        else:
            self._type_label = None

    def reorder(self, order, check_order=True):
        """Reorder the connections in place.

        Parameters
        ----------
        order : list
            New order of indexes.
        check_order : bool
            If ``True``, validate the new ``order`` before applying it.

        """
        # sanity check the sorting order before applying it
        if check_order and self.N > 1:
            sorted_order = numpy.sort(order)
            if (
                sorted_order[0] != 0
                or sorted_order[-1] != self.N - 1
                or not numpy.all(sorted_order[1:] == sorted_order[:-1] + 1)
            ):
                raise ValueError("New order must be an array from 0 to N-1")

        if self.has_id():
            self._id = self._id[order]
        if self.has_typeid():
            self._typeid = self._typeid[order]
        if self.has_members():
            self._members = self._members[order]


class Bonds(Topology):
    """Bond connections between particles.

    All values of indexes follow the LAMMPS 1-indexed convention, but the
    arrays themselves are 0-indexed.

    Parameters
    ----------
    N : int
        Number of bonds.
    num_types : int
        Number of bond types. Default of ``None`` means the number of types is
        determined from the unique typeids.

    Example
    -------
    Create bonds:

    .. code-block:: python

        bonds = lammpsio.topology.Bonds(N=3, num_types=1)
        bonds.id = [1, 2, 3]
        bonds.typeid = [1, 1, 1]
        bonds.members = [[1, 2], [2, 3], [1, 3]]

    This creates a molecule with three bonds of the same type.
    The bonds are defined between particle IDs 1-2, 2-3, and 1-3 to form a triangle
    structure.

    """

    def __init__(self, N: int, num_types: Optional[int] = None) -> None:
        super().__init__(N=N, num_members=2, num_types=num_types)


class Angles(Topology):
    """Angle connections between particles.

    All values of indexes follow the LAMMPS 1-indexed convention, but the
    arrays themselves are 0-indexed.

    Parameters
    ----------
    N : int
        Number of angles.
    num_types : int
        Number of angle types. Default of ``None`` means the number of types is
        determined from the unique typeids.

    Example
    -------
    Create angles:

    .. code-block:: python

        angles = lammpsio.topology.Angles(N=3, num_types=1)
        angles.id = [1, 2, 3]
        angles.typeid = [1, 1, 1]
        angles.members = [[1, 2, 3], [2, 3, 1], [3, 1, 2]]

    This creates a molecule with three angles of the same type.
    The angles are defined between particles IDs 1-2-3, 2-3-1, and 3-1-2
    to form a triangle structure.

    """

    def __init__(self, N: int, num_types: Optional[int] = None) -> None:
        super().__init__(N=N, num_members=3, num_types=num_types)


class Dihedrals(Topology):
    """Dihedral connections between particles.

    All values of indexes follow the LAMMPS 1-indexed convention, but the
    arrays themselves are 0-indexed.

    Parameters
    ----------
    N : int
        Number of dihedrals.
    num_types : int
        Number of dihedral types. Default of ``None`` means the number of types
        is determined from the unique typeids.

    Example
    -------
    Create dihedrals:

    .. code-block:: python

        dihedrals = lammpsio.topology.Dihedrals(N=2, num_types=2)
        dihedrals.id = [1, 2]
        dihedrals.typeid = [1, 2]
        dihedrals.members = [
            [1, 2, 3, 4],
            [2, 3, 4, 5],
        ]

    This creates two dihedrals of two different types for a molecule
    consisting of eight atoms.

    """

    def __init__(self, N: int, num_types: Optional[int] = None) -> None:
        super().__init__(N=N, num_members=4, num_types=num_types)


class Impropers(Topology):
    """Improper dihedral connections between particles.

    All values of indexes follow the LAMMPS 1-indexed convention, but the
    arrays themselves are 0-indexed.

    Parameters
    ----------
    N : int
        Number of improper dihedrals.
    num_types : int
        Number of improper dihedral types. Default of ``None`` means the number
        of types is determined from the unique typeids.

    Example
    -------
    Create improper dihedrals:

    .. code-block:: python

        impropers = lammpsio.topology.Impropers(N=2, num_types=2)
        impropers.id = [1, 2]
        impropers.typeid = [1, 2]
        impropers.members = [
            [1, 2, 3, 4],
            [2, 3, 4, 5],
        ]

    This creates two improper dihedrals of two different types for a molecule
    consisting of five atoms.

    """

    def __init__(self, N: int, num_types: Optional[int] = None) -> None:
        super().__init__(N=N, num_members=4, num_types=num_types)


class LabelMap(collections.abc.MutableMapping[int, str]):
    """Map between integer type IDs and string type names.

    A `LabelMap` is effectively a dictionary associating a label (type) with a
    particle's or connection's typeid. These labels can be useful for tracking
    the meaning of typeids. They are also automatically used when interconverting
    with HOOMD GSD files that require such labels.

    Parameters
    ----------
    map : dict
        Map of typeids to types.

    Example
    -------
    Create `LabelMap`:

    .. code-block:: python

        type_label = lammpsio.topology.LabelMap({1: "A", 2: "B"})

    This creates a dictionary mapping numeric type ID labels 1 and 2 used by LAMMPS
    to alphanumeric type labels "A" and "B", such as those used by HOOMD-blue. The
    map can be accessed directly:

    .. code-block:: python

        assert type_label[1] == "A"

    The `LabelMap` additionally supports inverse mapping from type label to type ID:

    .. code-block:: python

        assert type_label.inverse["B"] == 2
    """

    def __init__(self, map=None):
        self._map = {}
        self._inverse_map = {}
        if map is not None:
            self.update(map)

    def __getitem__(self, key):
        return self._map[key]

    def __setitem__(self, key, value):
        self._map[key] = value
        self._inverse_map[value] = key

    def __delitem__(self, key):
        value = self._map.pop(key)
        del self._inverse_map[value]

    def __iter__(self):
        return iter(self._map)

    def __len__(self):
        return len(self._map)

    @property
    def types(self):
        """tuple of str: Types in map."""
        return tuple(self._map.values())

    @property
    def typeid(self):
        """tuple of int: Type IDs in map."""
        return tuple(self._map.keys())

    @property
    def inverse(self) -> dict[str, int]:
        """dict: Inverse map from type label to type id."""
        return self._inverse_map



import gzip
import os
import pathlib
from typing import Any, Dict, List, Optional, Union

import numpy

from . import _compatibility
from .box import Box
from .data import _readline
from .snapshot import Snapshot

if _compatibility.pyzstd_version is not None:
    import pyzstd


class DumpFile:
    """LAMMPS dump file.

    The LAMMPS dump file is a highly flexible file format. The ``schema`` of the
    file is deduced from the ``ITEM: ATOMS`` header unless one is manually
    specified. If specificed, the ``schema`` must be a dictionary mapping pieces
    of data to column indexes (0-indexed). Valid keys for the schema match the
    names and shapes in the `Snapshot`. The keys requiring only 1 column index
    are:

    - ``"id"``
    - ``"typeid"``
    - ``"molecule"``
    - ``"charge"``
    - ``"mass"``

    The keys requiring 3 column indexes are:

    - ``"position"``
    - ``"velocity"``
    - ``"image"``

    LAMMPS dumps particles in an unknown order unless you have used the
    ``dump_modify sort`` option. If you want particles to be ordered by ``id``
    in the `Snapshot`, use ``sort_ids=True`` (default). Note that slightly
    faster reading may be possible by setting this option to ``False``.

    A `DumpFile` is iterable, so you can use it to go through all the snapshots
    of a trajectory. It supports LAMMPS dump file with different compression
    formats such as gzip and Zstandard
    (`pyzstd <https://pyzstd.readthedocs.io/en/stable/>`_ package needs to be installed
    separately for Zstandard support).
    Random access to snapshots is not currently implemented, but it may be added
    in future. If you want to randomly access snapshots, you should load the whole
    file into a list, but the memory requirements to do so may be large.

    The dump file may not contain certain information about your particles, for
    example, topology, or you may choose not to write this information in the
    dump file because it does not change frame-to-frame. The `copy_from` option
    allows this information to be copied into a new snapshot from a reference
    one, e.g, that was read from a `DataFile`.

    Parameters
    ----------
    filename : str
        Path to dump file.
    schema : dict
        Schema for the contents of the file. Defaults to ``None``, which means
        to read it from the file.
    sort_ids : bool
        If ``True``, sort the particles by ID in each snapshot.
    copy_from : `Snapshot`
        If specified, copy supported fields that are missing in the dump file
        but are set in a reference `Snapshot`.

    Example
    -------

    .. skip: start(lammps == None, reason="lammps not installed")

    .. invisible-code-block: python

        filename = tmp_path / "atoms.lammpstrj"
        lmps = lammps.lammps(cmdargs=["-log", "none", "-nocite"])

        cmds = []
        cmds += ["units lj"]
        cmds += ["atom_style atomic"]
        cmds += ["boundary p p p"]

        box_length = 3

        cmds += [f"region box block 0 {box_length} 0 {box_length} 0 {box_length}"]
        cmds += ["create_box 1 box"]

        cmds += ["create_atoms 1 random 10 12345 box"]
        cmds += ["mass 1 1.0"]

        cmds += ["pair_style lj/cut 2.5"]
        cmds += ["pair_coeff 1 1 1.0 1.0 2.5"]

        cmds += ["minimize 1.0e-4 1.0e-6 100 1000"]
        cmds += ["timestep 0.005"]
        cmds += ["fix 1 all nvt temp 1.0 1.0 0.1"]
        cmds += ["thermo 1000"]
        cmds += ["thermo_style custom step temp pe ke etotal press density"]
        cmds += [f"dump 1 all custom 100 {filename} id x y z ix iy iz"]
        cmds += ["run 1000"]

        lmps.commands_list(cmds)
        lmps.close()

    Create a dump file object:

    .. code-block:: python

        traj = lammpsio.DumpFile(filename)

    Iterate snapshots:

    .. code-block:: python

        for snapshot in traj:
            print(snapshot.step)

    You can also use multiprocessing to read snapshots in parallel:

    .. code-block:: python

        import multiprocessing

        def process_snapshot(snapshot):
            return snapshot.N

        if __name__ == '__main__':
            num_workers = max(2, multiprocessing.cpu_count())
            with multiprocessing.Pool(num_workers) as p:
                num_particles = p.map(process_snapshot, traj)
            print(num_particles)

    You can also get the number of snapshots in the `DumpFile`, but this does
    require reading the entire file: use with caution!

    .. code-block:: python

        num_frames = len(traj)

    Random access by creating a list:


    .. code-block:: python

        snapshots = [snap for snap in traj]
        print(snapshots[3].step)

    """

    def __init__(
        self,
        filename: str,
        schema: Optional[Dict] = None,
        sort_ids: bool = True,
        copy_from: Optional["Snapshot"] = None,
    ) -> None:
        self.filename = filename
        self.schema = schema
        self._frames = None
        self.sort_ids = sort_ids
        self.copy_from = copy_from

    @classmethod
    def create(
        cls: type["DumpFile"],
        filename: str,
        schema: Dict,
        snapshots: Union["Snapshot", List["Snapshot"]],
    ) -> "DumpFile":
        """Create a LAMMPS dump file.

        Parameters
        ----------
        filename : str
            Path to dump file.
        schema : dict
            Schema for the contents of the file.
        snapshots : `Snapshot` or list
            One or more snapshots to write to the dump file.

        Returns
        -------
        `DumpFile`
            The object representing the new dump file.

        Example
        -------
        A `DumpFile` can be created from a list of snapshots:

        .. code-block:: python

            schema = {"id":0, "position":(1, 2, 3), "image": (4, 5, 6)}

            lammpsio.DumpFile.create(filename, schema, snapshots)

        .. skip: end

        """
        # map out the schema into a dump row
        # each entry is a tuple: (column, (attribute, index))
        # the index is None for scalars, otherwise it is the component of the vector
        dump_row = []
        for k, v in schema.items():
            try:
                cols = iter(v)
                for i, col in enumerate(cols):
                    dump_row.append((col, (k, i)))
            except TypeError:
                dump_row.append((v, (k, None)))
        dump_row.sort(key=lambda x: x[0])

        # make snapshots iterable
        try:
            snapshots = iter(snapshots)
        except TypeError:
            snapshots = [snapshots]

        with open(filename, "w") as f:
            for snap in snapshots:
                f.write("ITEM: TIMESTEP\n" f"{snap.step}\n")

                f.write("ITEM: NUMBER OF ATOMS\n")
                f.write(f"{snap.N}\n")

                # always assume periodic in all directions
                if snap.box.tilt is not None:
                    f.write("ITEM: BOX BOUNDS xy xz yz pp pp pp\n")
                    xy, xz, yz = snap.box.tilt
                    lo = [
                        snap.box.low[0] + min([0.0, xy, xz, xy + xz]),
                        snap.box.low[1] + min([0.0, yz]),
                        snap.box.low[2],
                    ]
                    hi = [
                        snap.box.high[0] + max([0.0, xy, xz, xy + xz]),
                        snap.box.high[1] + max([0.0, yz]),
                        snap.box.high[2],
                    ]
                    for i in range(3):
                        f.write(f"{lo[i]:f} {hi[i]:f} {snap.box.tilt[i]:f}\n")
                else:
                    f.write("ITEM: BOX BOUNDS pp pp pp\n")
                    lo = snap.box.low
                    hi = snap.box.high
                    for i in range(3):
                        f.write(f"{lo[i]:f} {hi[i]:f}\n")

                # mapping from lammpsio to LAMMPS dump keys
                lammps_fields = {
                    "id": "id",
                    "molecule": "mol",
                    "typeid": "type",
                    "mass": "mass",
                    "position": ("x", "y", "z"),
                    "image": ("ix", "iy", "iz"),
                    "velocity": ("vx", "vy", "vz"),
                    "charge": "q",
                }
                schema_header = []
                for _, (key, key_idx) in dump_row:
                    field = lammps_fields[key]
                    if key_idx is not None:
                        field = field[key_idx]
                    schema_header.append(field)
                schema_header = " ".join(schema_header)

                f.write("ITEM: ATOMS " + schema_header + "\n")
                for i in range(snap.N):
                    line = ""
                    for col, (key, key_idx) in dump_row:
                        if key == "id":
                            val = snap.id[i] if snap.has_id() else i + 1
                        else:
                            val = getattr(snap, key)[i]
                            if key_idx is not None:
                                val = val[key_idx]

                        if key in ("id", "typeid", "molecule", "image"):
                            fmt = "{:d}"
                        elif key in ("position", "velocity"):
                            fmt = "{:.8f}"
                        else:
                            fmt = "{:f}"

                        if col > 0:
                            fmt = " " + fmt
                        line += fmt.format(val)
                    line = line.strip()
                    f.write(line + "\n")

        filename_path = pathlib.Path(filename)
        compression = cls._compression_from_suffix(filename_path.suffix)
        if compression:
            tmp = pathlib.Path(filename).with_suffix(filename_path.suffix + ".tmp")
            with open(filename, "rb") as src, compression.open(tmp, "wb") as dest:
                dest.writelines(src)
            os.replace(tmp, filename)

        return DumpFile(filename, schema)

    @staticmethod
    def _compression_from_suffix(suffix: str) -> Optional[Any]:
        if suffix == ".gz":
            return gzip
        elif suffix == ".zst":
            if _compatibility.pyzstd_version is None:
                raise ModuleNotFoundError("pyzstd needed for zstd compression")
            return pyzstd
        else:
            return None

    @property
    def copy_from(self) -> Optional["Snapshot"]:
        """`Snapshot`: Copy fields that are missing in a dump file.

        The fields that can be copied are:

        - `Snapshot.angles`
        - `Snapshot.bonds`
        - `Snapshot.charge`
        - `Snapshot.dihedrals`
        - `Snapshot.impropers`
        - `Snapshot.mass`
        - `Snapshot.molecule`
        - `Snapshot.typeid`

        """
        return self._copy_from

    @copy_from.setter
    def copy_from(self, value: Optional["Snapshot"]) -> None:
        if value is not None and not isinstance(value, Snapshot):
            raise TypeError("Dump file can only copy from Snapshot")
        self._copy_from = value

    @property
    def filename(self) -> str:
        """str: Path to the file."""
        return self._filename

    @filename.setter
    def filename(self, value: str) -> None:
        self._filename = value
        self._compression = self._compression_from_suffix(pathlib.Path(value).suffix)

        # configure section labels of dump file
        self._section = {
            "step": "ITEM: TIMESTEP",
            "natoms": "ITEM: NUMBER OF ATOMS",
            "box": "ITEM: BOX BOUNDS",
            "atoms": "ITEM: ATOMS",
        }
        if self._compression:
            for key, val in self._section.items():
                self._section[key] = val.encode()

    @property
    def schema(self) -> Optional[Dict]:
        """dict: Data schema."""
        return self._schema

    @schema.setter
    def schema(self, value: Optional[Dict]) -> None:
        if value is not None:
            # validate schema
            if "position" in value and len(value["position"]) != 3:
                raise ValueError("Position must be a 3-tuple")
            if "velocity" in value and len(value["velocity"]) != 3:
                raise ValueError("Velocity must be a 3-tuple")
            if "image" in value and len(value["image"]) != 3:
                raise ValueError("Image must be a 3-tuple")
        self._schema = value

    def _open(self):
        """Open the file handle for reading."""
        if self._compression:
            f = self._compression.open(self.filename, "rb")
        else:
            f = open(self.filename, "r")
        return f

    def _find_frames(self):
        """Seek line numbers for each frame."""
        self._frames = []
        with self._open() as f:
            line = _readline(f)
            line_num = 0
            while len(line) > 0:
                if self._section["step"] in line:
                    self._frames.append(line_num)
                line = _readline(f)
                line_num += 1

    def __len__(self):
        if self._frames is None:
            self._find_frames()
        return len(self._frames)

    def __iter__(self):
        with self._open() as f:
            state = 0
            line = _readline(f)
            while len(line) > 0:
                # timestep line first
                if state == 0 and self._section["step"] in line:
                    state += 1
                    step = int(_readline(f, True))

                # number of particles second
                if state == 1 and self._section["natoms"] in line:
                    state += 1
                    N = int(_readline(f, True))

                # box size third
                if state == 2 and self._section["box"] in line:
                    state += 1
                    box_header = line.split()
                    # check for triclinic
                    if len(box_header) == 9:
                        is_triclinic = True
                    elif len(box_header) == 6:
                        is_triclinic = False
                    else:
                        raise IOError("Incorrectly formed box bound header")
                    box_ = [
                        [float(v) for v in _readline(f, True).split()]
                        for line_ in range(3)
                    ]
                    x_lo, x_hi = box_[0][:2]
                    y_lo, y_hi = box_[1][:2]
                    z_lo, z_hi = box_[2][:2]
                    if is_triclinic:
                        xy, xz, yz = [row[2] for row in box_]
                        lo = [
                            x_lo - min([0.0, xy, xz, xy + xz]),
                            y_lo - min([0.0, yz]),
                            z_lo,
                        ]
                        hi = [
                            x_hi - max([0.0, xy, xz, xy + xz]),
                            y_hi - max([0.0, yz]),
                            z_hi,
                        ]
                        box = Box(lo, hi, [xy, xz, yz])
                    else:
                        box = Box([x_lo, y_lo, z_lo], [x_hi, y_hi, z_hi])

                # atoms come fourth
                if state == 3 and self._section["atoms"] in line:
                    state += 1

                    # extract the schema
                    if self.schema is None:
                        # mapping from LAMMPS dump keys to lammpsio
                        lammps_fields = {
                            "id": ("id", None),
                            "mol": ("molecule", None),
                            "type": ("typeid", None),
                            "mass": ("mass", None),
                            "x": ("position", 0),
                            "y": ("position", 1),
                            "z": ("position", 2),
                            "ix": ("image", 0),
                            "iy": ("image", 1),
                            "iz": ("image", 2),
                            "vx": ("velocity", 0),
                            "vy": ("velocity", 1),
                            "vz": ("velocity", 2),
                            "q": ("charge", None),
                        }
                        schema = {}
                        schema_header = line.split()[2:]
                        for i, field in enumerate(schema_header):
                            if self._compression:
                                field = field.decode()
                            if field in lammps_fields:
                                key, key_idx = lammps_fields[field]
                                if key_idx is not None:
                                    if key not in schema:
                                        schema[key] = [None, None, None]
                                    schema[key][key_idx] = i
                                else:
                                    schema[key] = i
                        # validate tuple
                        for key in ("position", "velocity", "image"):
                            if key in schema and any(x is None for x in schema[key]):
                                raise IOError("lammpsio requires 3-element vectors")
                        self.schema = schema

                    snap = Snapshot(N, box, step)
                    for i in range(snap.N):
                        atom = _readline(f, True)
                        atom = atom.split()

                        if "id" in self.schema:
                            id_ = int(atom[self.schema["id"]])
                            if id_ != i + 1:
                                snap.id[i] = id_
                        if "position" in self.schema:
                            snap.position[i] = [
                                float(atom[j]) for j in self.schema["position"]
                            ]
                        if "velocity" in self.schema:
                            snap.velocity[i] = [
                                float(atom[j]) for j in self.schema["velocity"]
                            ]
                        if "image" in self.schema:
                            snap.image[i] = [int(atom[j]) for j in self.schema["image"]]
                        if "molecule" in self.schema:
                            snap.molecule[i] = int(atom[self.schema["molecule"]])
                        if "typeid" in self.schema:
                            snap.typeid[i] = int(atom[self.schema["typeid"]])
                        if "charge" in self.schema:
                            snap.charge[i] = float(atom[self.schema["charge"]])
                        if "mass" in self.schema:
                            snap.mass[i] = float(atom[self.schema["mass"]])

                # final processing stage for the frame
                if state == 4:
                    # optionally sort the particles by ID
                    if self.sort_ids and snap.has_id():
                        snap.reorder(numpy.argsort(snap.id), check_order=False)

                    # optionally copy reference data by ID / index
                    if self._copy_from is not None:
                        if snap.N != self._copy_from.N:
                            raise ValueError(
                                "Cannot copy from a Snapshot with a different size"
                            )

                        if self._copy_from.has_id():
                            copy_id_map = {
                                id_: i for i, id_ in enumerate(self._copy_from.id)
                            }
                        else:
                            copy_id_map = {i + 1: i for i in range(self._copy_from.N)}

                        if snap.has_id():
                            copy_id = [copy_id_map[id_] for id_ in snap.id]
                        else:
                            copy_id = [copy_id_map[id_] for id_ in range(1, snap.N + 1)]

                        if not snap.has_typeid() and self._copy_from.has_typeid():
                            snap.typeid = self._copy_from.typeid[copy_id]
                        if not snap.has_molecule() and self._copy_from.has_molecule():
                            snap.molecule = self._copy_from.molecule[copy_id]
                        if not snap.has_charge() and self._copy_from.has_charge():
                            snap.charge = self._copy_from.charge[copy_id]
                        if not snap.has_mass() and self._copy_from.has_mass():
                            snap.mass = self._copy_from.mass[copy_id]
                        if self._copy_from.has_bonds():
                            snap.bonds = self._copy_from.bonds
                        if self._copy_from.has_angles():
                            snap.angles = self._copy_from.angles
                        if self._copy_from.has_dihedrals():
                            snap.dihedrals = self._copy_from.dihedrals
                        if self._copy_from.has_impropers():
                            snap.impropers = self._copy_from.impropers

                    yield snap
                    del snap, N, box, step
                    state = 0

                line = _readline(f)



import numpy
import pytest

import lammpsio

try:
    import lammps

    has_lammps = True
except ImportError:
    has_lammps = False


@pytest.mark.parametrize("shuffle_ids", [False, True])
@pytest.mark.parametrize("atom_style", ["atomic", "molecular", "charge", "full"])
def test_data_file_min(snap, atom_style, shuffle_ids, tmp_path):
    if shuffle_ids:
        snap.id = [2, 0, 1]
    # write the data file with default values
    filename = tmp_path / "atoms.data"
    data = lammpsio.DataFile.create(filename, snap, atom_style)
    assert filename.exists

    # read it back in and check
    snap_2 = data.read()
    assert snap_2.N == snap.N
    assert snap_2.step is None
    assert numpy.allclose(snap_2.box.low, snap.box.low)
    assert numpy.allclose(snap_2.box.high, snap.box.high)
    if snap.box.tilt is not None:
        assert numpy.allclose(snap_2.box.tilt, snap.box.tilt)
    else:
        assert snap_2.box.tilt is None
    if shuffle_ids:
        assert snap_2.has_id()
        assert numpy.allclose(snap_2.id, snap.id)
    else:
        assert not snap_2.has_id()
    assert snap_2.has_position()
    assert numpy.allclose(snap_2.position, 0)
    assert not snap_2.has_image()
    assert not snap_2.has_velocity()
    assert snap_2.has_typeid()
    assert numpy.allclose(snap_2.typeid, 1)
    assert not snap_2.has_mass()
    if atom_style in ("molecular", "full"):
        assert snap_2.has_molecule()
        assert numpy.allclose(snap_2.molecule, 0)
    else:
        assert not snap_2.has_molecule()
    if atom_style in ("charge", "full"):
        assert snap_2.has_charge()
        assert numpy.allclose(snap_2.charge, 0)
    else:
        assert not snap_2.has_charge()


@pytest.mark.parametrize("shuffle_ids", [False, True])
@pytest.mark.parametrize("set_style", [True, False])
@pytest.mark.parametrize("atom_style", ["atomic", "molecular", "charge", "full"])
def test_data_file_all(snap, atom_style, set_style, shuffle_ids, tmp_path):
    if shuffle_ids:
        snap.id = [2, 0, 1]
    # write the data file with nondefault values
    snap.position = [[0.1, 0.2, 0.3], [-0.4, -0.5, -0.6], [0.7, 0.8, 0.9]]
    snap.image = [[1, 2, 3], [-4, -5, -6], [7, 8, 9]]
    snap.velocity = [[-3, -2, -1], [6, 5, 4], [9, 8, 7]]
    snap.typeid = [2, 1, 2]
    snap.mass = [3, 2, 3]
    if atom_style in ("molecular", "full"):
        snap.molecule = [2, 0, 1]
    if atom_style in ("charge", "full"):
        snap.charge = [-1, 0, 1]
    filename = tmp_path / "atoms.data"
    data = lammpsio.DataFile.create(filename, snap, atom_style if set_style else None)
    assert filename.exists

    # read it back in and check
    snap_2 = data.read()
    assert snap_2.N == snap.N
    assert snap_2.step is None
    assert numpy.allclose(snap_2.box.low, snap.box.low)
    assert numpy.allclose(snap_2.box.high, snap.box.high)
    if snap.box.tilt is not None:
        assert numpy.allclose(snap_2.box.tilt, snap.box.tilt)
    else:
        assert snap_2.box.tilt is None
    if shuffle_ids:
        assert snap_2.has_id()
        assert numpy.allclose(snap_2.id, snap.id)
    else:
        assert not snap_2.has_id()
    assert snap_2.has_position()
    assert numpy.allclose(snap_2.position, snap.position)
    assert snap_2.has_velocity()
    assert numpy.allclose(snap_2.velocity, snap.velocity)
    assert snap_2.has_image()
    assert numpy.allclose(snap_2.image, snap.image)
    assert snap_2.has_typeid()
    assert numpy.allclose(snap_2.typeid, snap.typeid)
    assert snap_2.has_mass()
    assert numpy.allclose(snap_2.mass, snap.mass)
    if atom_style in ("molecular", "full"):
        assert snap_2.has_molecule()
        assert numpy.allclose(snap_2.molecule, snap.molecule)
    else:
        assert not snap_2.has_molecule()
    if atom_style in ("charge", "full"):
        assert snap_2.has_charge()
        assert numpy.allclose(snap_2.charge, snap.charge)
    else:
        assert not snap_2.has_charge()


@pytest.mark.parametrize("shuffle_ids", [False, True])
def test_data_file_topology(snap_8, tmp_path, shuffle_ids):
    # set ids to be assigned
    if shuffle_ids:
        particle_id = [1, 5, 2, 6, 3, 7, 4, 8]
        bond_id = [1, 4, 2, 5, 3, 6]
        angle_id = [1, 3, 2, 4]
        dihedral_id = [2, 1]
        improper_id = [2, 1]
    else:
        particle_id = [1, 2, 3, 4, 5, 6, 7, 8]
        bond_id = [1, 2, 3, 4, 5, 6]
        angle_id = [1, 2, 3, 4]
        dihedral_id = [1, 2]
        improper_id = [1, 2]

    # particle information
    snap_8.id = particle_id
    snap_8.typeid = [1, 1, 1, 1, 2, 2, 2, 2]
    snap_8.position = [
        [0, 0, 0],
        [0.1, 0.1, 0.1],
        [0.2, 0.2, 0.2],
        [0.3, 0.3, 0.3],
        [1, 1, 1],
        [1.1, 1.1, 1.1],
        [1.2, 1.2, 1.2],
        [1.3, 1.3, 1.3],
    ]
    snap_8.mass = [1, 1, 1, 1, 2, 2, 2, 2]

    # bond information
    snap_8.bonds = lammpsio.topology.Bonds(N=6, num_types=2)
    snap_8.bonds.id = bond_id
    snap_8.bonds.typeid = [1, 2, 1, 2, 1, 2]
    snap_8.bonds.members = [
        [1, 2],
        [2, 3],
        [3, 4],
        [5, 6],
        [6, 7],
        [7, 8],
    ]

    # angle information
    snap_8.angles = lammpsio.topology.Angles(N=4, num_types=2)
    snap_8.angles.id = angle_id
    snap_8.angles.typeid = [1, 2, 2, 1]
    snap_8.angles.members = [
        [1, 2, 3],
        [2, 3, 4],
        [5, 6, 7],
        [6, 7, 8],
    ]

    # dihedral information
    snap_8.dihedrals = lammpsio.topology.Dihedrals(N=2, num_types=2)
    snap_8.dihedrals.id = dihedral_id
    snap_8.dihedrals.typeid = [1, 2]
    snap_8.dihedrals.members = [
        [1, 2, 3, 4],
        [5, 6, 7, 8],
    ]

    # improper information
    snap_8.impropers = lammpsio.topology.Impropers(N=2, num_types=2)
    snap_8.impropers.id = improper_id
    snap_8.impropers.typeid = [1, 2]
    snap_8.impropers.members = [
        [1, 2, 3, 4],
        [5, 6, 7, 8],
    ]

    filename = tmp_path / "atoms.data"
    data = lammpsio.DataFile.create(filename, snap_8)
    assert filename.exists

    snap_2 = data.read()
    # test non-topology features of a snapshot with topology
    assert snap_2.N == snap_8.N
    if shuffle_ids:
        assert snap_2.has_id()
        assert numpy.allclose(snap_2.id, snap_8.id)
    else:
        assert not snap_2.has_id()
    assert numpy.allclose(snap_2.id, snap_8.id)
    assert snap_2.has_typeid()
    assert numpy.allclose(snap_2.typeid, snap_8.typeid)
    assert snap_2.has_position()
    assert numpy.allclose(snap_2.position, snap_8.position)
    assert snap_2.has_mass()
    assert numpy.allclose(snap_2.mass, snap_8.mass)

    # test bonds
    assert snap_2.bonds.N == snap_8.bonds.N
    if shuffle_ids:
        assert snap_2.bonds.has_id()
        assert numpy.allclose(snap_2.bonds.id, snap_8.bonds.id)
    else:
        assert not snap_2.bonds.has_id()
    assert numpy.allclose(snap_2.bonds.id, snap_8.bonds.id)
    assert snap_2.bonds.has_typeid()
    assert numpy.allclose(snap_2.bonds.typeid, snap_8.bonds.typeid)
    assert snap_2.bonds.has_members()
    assert numpy.allclose(snap_2.bonds.members, snap_8.bonds.members)

    # test angles
    assert snap_2.angles.N == snap_8.angles.N
    if shuffle_ids:
        assert snap_2.angles.has_id()
        assert numpy.allclose(snap_2.angles.id, snap_8.angles.id)
    else:
        assert not snap_2.angles.has_id()
    assert numpy.allclose(snap_2.angles.id, snap_8.angles.id)
    assert snap_2.angles.has_typeid()
    assert numpy.allclose(snap_2.angles.typeid, snap_8.angles.typeid)
    assert snap_2.angles.has_members()
    assert numpy.allclose(snap_2.angles.members, snap_8.angles.members)

    # test dihedrals
    assert snap_2.dihedrals.N == snap_8.dihedrals.N
    if shuffle_ids:
        assert snap_2.dihedrals.has_id()
        assert numpy.allclose(snap_2.dihedrals.id, snap_8.dihedrals.id)
    else:
        assert not snap_2.dihedrals.has_id()
    assert numpy.allclose(snap_2.dihedrals.id, snap_8.dihedrals.id)
    assert snap_2.dihedrals.has_typeid()
    assert numpy.allclose(snap_2.dihedrals.typeid, snap_8.dihedrals.typeid)
    assert snap_2.dihedrals.has_members()
    assert numpy.allclose(snap_2.dihedrals.members, snap_8.dihedrals.members)

    # test impropers
    assert snap_2.impropers.N == snap_8.impropers.N
    if shuffle_ids:
        assert snap_2.impropers.has_id()
        assert numpy.allclose(snap_2.impropers.id, snap_8.impropers.id)
    else:
        assert not snap_2.impropers.has_id()
    assert numpy.allclose(snap_2.impropers.id, snap_8.impropers.id)
    assert snap_2.impropers.has_typeid()
    assert numpy.allclose(snap_2.impropers.typeid, snap_8.impropers.typeid)
    assert snap_2.impropers.has_members()
    assert numpy.allclose(snap_2.impropers.members, snap_8.impropers.members)


@pytest.mark.lammps
@pytest.mark.skipif(not has_lammps, reason="lammps not installed")
@pytest.mark.parametrize("shuffle_ids", [False, True])
@pytest.mark.parametrize("atom_style", ["atomic", "molecular", "charge", "full"])
def test_data_file_min_lammps(tmp_path, snap, atom_style, shuffle_ids):
    if shuffle_ids:
        snap.id = [3, 1, 2]

    # write the data file with default values using lammpsio
    filename = tmp_path / "atoms.data"
    assert not snap.has_image()
    assert not snap.has_velocity()
    lammpsio.DataFile.create(filename, snap, atom_style)
    assert filename.exists

    # read it in LAMMPS and write it out
    lmps = lammps.lammps(cmdargs=["-log", "none", "-nocite"])
    cmds = [f"atom_style {atom_style}"]
    cmds += [f"read_data {filename}"]
    cmds += ["mass 1 1.0"]
    cmds += [f"write_data {filename} nocoeff nofix"]
    lmps.commands_list(cmds)
    lmps.close()

    # read it back in and check
    snap_2 = lammpsio.DataFile(filename).read()
    assert snap_2.N == snap.N
    assert snap_2.step is None
    assert numpy.allclose(snap_2.box.low, snap.box.low)
    assert numpy.allclose(snap_2.box.high, snap.box.high)
    if snap.box.tilt is not None:
        assert numpy.allclose(snap_2.box.tilt, snap.box.tilt)
    else:
        assert snap_2.box.tilt is None
    if shuffle_ids:
        assert snap_2.has_id()
        assert numpy.allclose(snap_2.id, snap.id)
    else:
        assert not snap_2.has_id()
    assert snap_2.has_position()
    assert numpy.allclose(snap_2.position, 0)
    assert snap_2.has_typeid()
    assert numpy.allclose(snap_2.typeid, 1)

    if atom_style in ("molecular", "full"):
        assert snap_2.has_molecule()
        assert numpy.allclose(snap_2.molecule, 0)
    else:
        assert not snap_2.has_molecule()
    if atom_style in ("charge", "full"):
        assert snap_2.has_charge()
        assert numpy.allclose(snap_2.charge, 0)
    else:
        assert not snap_2.has_charge()


@pytest.mark.lammps
@pytest.mark.skipif(not has_lammps, reason="lammps not installed")
@pytest.mark.parametrize("shuffle_ids", [False, True])
@pytest.mark.parametrize("set_style", [True, False])
@pytest.mark.parametrize("atom_style", ["atomic", "molecular", "charge", "full"])
def test_data_file_all_lammps(snap, atom_style, set_style, shuffle_ids, tmp_path):
    if shuffle_ids:
        snap.id = [3, 1, 2]

    # write the data file with nondefault values using lammpsio
    snap.position = [[0.1, 0.2, 0.3], [-0.4, -0.5, -0.6], [0.7, 0.8, 0.9]]
    snap.image = [[1, 2, 3], [-4, -5, -6], [7, 8, 9]]
    snap.velocity = [[-3, -2, -1], [6, 5, 4], [9, 8, 7]]
    snap.typeid = [2, 1, 2]
    snap.mass = [3, 2, 3]
    if atom_style in ("molecular", "full"):
        snap.molecule = [2, 0, 1]
    if atom_style in ("charge", "full"):
        snap.charge = [-1, 0, 1]

    filename = tmp_path / "atoms.data"

    # create the data file using lammpsio
    lammpsio.DataFile.create(filename, snap, atom_style if set_style else None)
    assert filename.exists

    lmps = lammps.lammps(cmdargs=["-log", "none", "-nocite"])
    cmds = [f"atom_style {atom_style}"]
    cmds += [f"read_data {filename}"]
    cmds += [f"write_data {filename}"]
    lmps.commands_list(cmds)
    lmps.close()

    # read it back in and check
    snap_2 = lammpsio.DataFile(filename).read()
    assert snap_2.N == snap.N
    assert snap_2.step is None
    assert numpy.allclose(snap_2.box.low, snap.box.low)
    assert numpy.allclose(snap_2.box.high, snap.box.high)
    if snap.box.tilt is not None:
        assert numpy.allclose(snap_2.box.tilt, snap.box.tilt)
    else:
        assert snap_2.box.tilt is None
    if shuffle_ids:
        assert snap_2.has_id()
        assert numpy.allclose(snap_2.id, snap.id)
    else:
        assert not snap_2.has_id()
    assert snap_2.has_position()
    assert numpy.allclose(snap_2.position, snap.position)
    assert snap_2.has_velocity()
    assert numpy.allclose(snap_2.velocity, snap.velocity)
    assert snap_2.has_image()
    assert numpy.allclose(snap_2.image, snap.image)
    assert snap_2.has_typeid()
    assert numpy.allclose(snap_2.typeid, snap.typeid)
    assert snap_2.has_mass()
    assert numpy.allclose(snap_2.mass, snap.mass)
    if atom_style in ("molecular", "full"):
        assert snap_2.has_molecule()
        assert numpy.allclose(snap_2.molecule, snap.molecule)
    else:
        assert not snap_2.has_molecule()
    if atom_style in ("charge", "full"):
        assert snap_2.has_charge()
        assert numpy.allclose(snap_2.charge, snap.charge)
    else:
        assert not snap_2.has_charge()


@pytest.mark.lammps
@pytest.mark.skipif(not has_lammps, reason="lammps not installed")
@pytest.mark.parametrize("shuffle_ids", [False, True])
def test_data_file_topology_lammps(snap_8, tmp_path, shuffle_ids):
    # set ids to be assigned
    # Note: LAMMPS reorders the ids of topology objects when ids are shuffled
    # We only check that topology objects typeid and members survive the round
    # trip, but in future, we could check the IDs if this behavior were changed.
    if shuffle_ids:
        particle_id = [1, 5, 2, 6, 3, 7, 4, 8]
        bond_id = [1, 4, 2, 5, 3, 6]
        angle_id = [1, 3, 2, 4]
        dihedral_id = [2, 1]
        improper_id = [2, 1]
    else:
        particle_id = [1, 2, 3, 4, 5, 6, 7, 8]
        bond_id = [1, 2, 3, 4, 5, 6]
        angle_id = [1, 2, 3, 4]
        dihedral_id = [1, 2]
        improper_id = [1, 2]

    # particle information
    snap_8.id = particle_id
    snap_8.typeid = [1, 1, 1, 1, 2, 2, 2, 2]
    snap_8.position = [
        [0, 0, 0],
        [0.1, 0.1, 0.1],
        [0.2, 0.2, 0.2],
        [0.3, 0.3, 0.3],
        [1, 1, 1],
        [1.1, 1.1, 1.1],
        [1.2, 1.2, 1.2],
        [1.3, 1.3, 1.3],
    ]
    snap_8.mass = [1, 1, 1, 1, 2, 2, 2, 2]

    # bond information
    snap_8.bonds = lammpsio.topology.Bonds(N=6, num_types=2)
    snap_8.bonds.id = bond_id
    snap_8.bonds.typeid = [1, 2, 1, 2, 1, 2]
    snap_8.bonds.members = [
        [1, 2],
        [2, 3],
        [3, 4],
        [5, 6],
        [6, 7],
        [7, 8],
    ]

    # angle information
    snap_8.angles = lammpsio.topology.Angles(N=4, num_types=2)
    snap_8.angles.id = angle_id
    snap_8.angles.typeid = [1, 2, 2, 1]
    snap_8.angles.members = [
        [1, 2, 3],
        [2, 3, 4],
        [5, 6, 7],
        [6, 7, 8],
    ]

    # dihedral information
    snap_8.dihedrals = lammpsio.topology.Dihedrals(N=2, num_types=2)
    snap_8.dihedrals.id = dihedral_id
    snap_8.dihedrals.typeid = [1, 2]
    snap_8.dihedrals.members = [
        [1, 2, 3, 4],
        [5, 6, 7, 8],
    ]

    # improper information
    snap_8.impropers = lammpsio.topology.Impropers(N=2, num_types=2)
    snap_8.impropers.id = improper_id
    snap_8.impropers.typeid = [1, 2]
    snap_8.impropers.members = [
        [1, 2, 3, 4],
        [5, 6, 7, 8],
    ]

    filename = tmp_path / "atoms.data"
    # create the data file using lammpsio
    lammpsio.DataFile.create(filename, snap_8)

    # read it in LAMMPS and write it out
    lmps = lammps.lammps(cmdargs=["-log", "none", "-nocite"])
    cmds = ["atom_style molecular"]
    cmds += [f"read_data {filename}"]
    cmds += [f"write_data {filename}"]
    lmps.commands_list(cmds)
    lmps.close()

    # read it back in and check
    snap_2 = lammpsio.DataFile(filename).read()
    # test non-topology features of a snapshot with topology
    assert snap_2.N == snap_8.N
    if shuffle_ids:
        assert snap_2.has_id()
        assert numpy.allclose(snap_2.id, snap_8.id)
    else:
        assert not snap_2.has_id()

    assert numpy.allclose(snap_2.id, snap_8.id)
    assert snap_2.has_typeid()
    assert numpy.allclose(snap_2.typeid, snap_8.typeid)
    assert snap_2.has_position()
    assert numpy.allclose(snap_2.position, snap_8.position)
    assert snap_2.has_mass()
    assert numpy.allclose(snap_2.mass, snap_8.mass)

    # test bonds
    assert snap_2.bonds.N == snap_8.bonds.N
    assert snap_2.bonds.has_typeid()
    assert snap_2.bonds.has_members()

    # test that types and bond members survive round trip
    # Note: This sorting only works because each member only appears in the first
    # column one time.
    snap_8_bond_data = numpy.column_stack((snap_8.bonds.members, snap_8.bonds.typeid))
    snap_2_bond_data = numpy.column_stack((snap_2.bonds.members, snap_2.bonds.typeid))
    assert numpy.allclose(
        snap_2_bond_data[snap_2_bond_data[:, 0].argsort()],
        snap_8_bond_data[snap_8_bond_data[:, 0].argsort()],
    )

    # test angles
    assert snap_2.angles.N == snap_8.angles.N
    assert snap_2.angles.has_typeid()
    assert snap_2.angles.has_members()

    # test that types and angle members survive round trip
    snap_8_angle_data = numpy.column_stack(
        (snap_8.angles.members, snap_8.angles.typeid)
    )
    snap_2_angle_data = numpy.column_stack(
        (snap_2.angles.members, snap_2.angles.typeid)
    )
    assert numpy.allclose(
        snap_2_angle_data[snap_2_angle_data[:, 0].argsort()],
        snap_8_angle_data[snap_8_angle_data[:, 0].argsort()],
    )

    # test dihedrals
    assert snap_2.dihedrals.N == snap_8.dihedrals.N
    assert snap_2.dihedrals.has_typeid()
    assert snap_2.dihedrals.has_members()

    # test that types and dihedral members survive round trip
    snap_8_dihedral_data = numpy.column_stack(
        (snap_8.dihedrals.members, snap_8.dihedrals.typeid)
    )
    snap_2_dihedral_data = numpy.column_stack(
        (snap_2.dihedrals.members, snap_2.dihedrals.typeid)
    )
    assert numpy.allclose(
        snap_2_dihedral_data[snap_2_dihedral_data[:, 0].argsort()],
        snap_8_dihedral_data[snap_8_dihedral_data[:, 0].argsort()],
    )

    # test impropers
    assert snap_2.impropers.N == snap_8.impropers.N
    assert snap_2.impropers.has_typeid()
    assert snap_2.impropers.has_members()

    # test that types and improper members survive round trip
    snap_8_improper_data = numpy.column_stack(
        (snap_8.impropers.members, snap_8.impropers.typeid)
    )
    snap_2_improper_data = numpy.column_stack(
        (snap_2.impropers.members, snap_2.impropers.typeid)
    )
    assert numpy.allclose(
        snap_2_improper_data[snap_2_improper_data[:, 0].argsort()],
        snap_8_improper_data[snap_8_improper_data[:, 0].argsort()],
    )
